"""Utilities for generating image embeddings and inserting them into a
VectorDB database.

This script provides functions for:

    - Extracting payloads from LuxonisDatasets, specifically for classification datasets.
    - Filtering new samples based on their instance IDs to avoid duplicates in the database.
    - Generating embeddings for new images using an ONNX runtime session.
    - Performing batch upserts of embeddings into a VectorDB database.

Key modules used:

    - I{luxonis_ml.data}: For loading and working with LuxonisDatasets.
    - I{luxonis_ml.embeddings.utils.embedding}: For extracting embeddings from images.
    - I{luxonis_ml.embeddings.utils.vectordb}: For interacting with VectorDB databases.

Main functions:

    - I{_get_sample_payloads_LDF}: Extracts payloads from a LuxonisDataset for classification datasets.
    - I{_filter_new_samples_by_id}: Filters out samples already in the database based on instance IDs.
    - I{_batch_upsert}: Performs batch upserts of embeddings into the database.
    - I{generate_embeddings}: Main function to generate embeddings for a dataset and insert them into the database.

Important note:

Ensure that a VectorDB server is running and accessible before using these utilities.
"""

from typing import Any, Callable, Dict, List

import numpy as np
import onnxruntime

from luxonis_ml.data import LuxonisDataset
from luxonis_ml.embeddings.utils.embedding import extract_embeddings
from luxonis_ml.embeddings.utils.vectordb import VectorDBAPI


def _get_sample_payloads_LDF(dataset: LuxonisDataset) -> List[Dict[str, Any]]:
    """Extract payloads from the LuxonisDataset. Currently supports
    classification datasets.

    @type dataset: LuxonisDataset
    @param dataset: An instance of LuxonisDataset.
    @rtype: List[Dict[str, Any]]
    @return: List of payloads.
    """

    all_payloads = []
    df = dataset._load_df_offline()
    index = dataset._get_index()

    df = df.merge(index, on="instance_id")

    for row in df.iterrows():
        if row[1]["type"] == "classification":
            instance_id = row[1]["instance_id"]
            img_path = row[1]["original_filepath"]
            class_name = row[1]["class"]

            all_payloads.append(
                {
                    "instance_id": instance_id,
                    "image_path": img_path,
                    "class": class_name,
                }
            )

    return all_payloads


def _filter_new_samples_by_id(
    vectordb_api: VectorDBAPI, all_payloads: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """Filter out samples that are already in the Vector database based
    on their instance ID.

    @type vectordb_api: L{VectorDBAPI}
    @param vectordb_api: Vector database API instance.
    @type all_payloads: List[Dict[str, Any]]
    @param all_payloads: List of all payloads.
    @rtype: List[Dict[str, Any]]
    @return: List of new payloads.
    """
    all_payloads = all_payloads or []
    # Filter out samples that are already in the Qdrant database
    if len(all_payloads) == 0:
        print("Payloads list is empty!")
        return all_payloads

    retrieved_ids = vectordb_api.retrieve_all_ids()

    new_payloads = [
        payload
        for payload in all_payloads
        if payload["instance_id"] not in retrieved_ids
    ]

    return new_payloads


def _batch_upsert(
    vectordb_api: VectorDBAPI,
    new_embeddings: List[List[float]],
    new_payloads: List[Dict[str, Any]],
    vectordb_batch_size: int = 64,
) -> None:
    """Perform batch upserts of embeddings to VectorDB.

    @type vectordb_api: L{VectorDBAPI}
    @param vectordb_api: VectorDBAPI instance.
    @type new_embeddings: List[List[float]]
    @param new_embeddings: List of new embeddings.
    @type new_payloads: List[Dict[str, Any]]
    @param new_payloads: List of new payloads.
    @type vectordb_batch_size: int
    @param vectordb_batch_size: Batch size for inserting into VectorDB.
    """
    uuids = []
    qdrant_payloads = []
    for payload in new_payloads:
        uuids.append(payload["instance_id"])
        qdrant_payloads.append(
            {"label": payload["class"], "image_path": payload["image_path"]}
        )

    try:
        vectordb_api.insert_embeddings(
            uuids, new_embeddings, qdrant_payloads, vectordb_batch_size
        )
        print(f"Upserted {len(uuids)} of embeddings to VectorDB.")

    except Exception as e:
        print(e)
        print("Failed to upsert embeddings to VectorDB.")


def generate_embeddings(
    luxonis_dataset: LuxonisDataset,
    ort_session: onnxruntime.InferenceSession,
    vectordb_api: VectorDBAPI,
    output_layer_name: str = "/Flatten_output_0",
    transform: Callable[[np.ndarray], np.ndarray] = None,
    emb_batch_size: int = 64,
    vectordb_batch_size: int = 64,
) -> Dict[str, List[float]]:
    """Generate embeddings for a given dataset and insert them into a
    VectorDB.

    @type luxonis_dataset: L{LuxonisDataset}
    @param luxonis_dataset: The dataset object.
    @type ort_session: L{InferenceSession}
    @param ort_session: ONNX runtime session.
    @type vectordb_api: L{VectorDBAPI}
    @param vectordb_api: VectorDBAPI instance.
    @type output_layer_name: str
    @param output_layer_name: Name of the output layer in the ONNX
        model.
    @type transform: Callable[[np.ndarray], np.ndarray]
    @param transform: Preprocessing function for images. If None,
        default preprocessing is used.
    @type emb_batch_size: int
    @param emb_batch_size: Batch size for generating embeddings.
    @type vectordb_batch_size: int
    @param vectordb_batch_size: Batch size for inserting into a vector
        DB.
    @type: Dict[str, List[float]]
    @return: Dictionary of instance ID to embedding.
    """

    all_payloads = _get_sample_payloads_LDF(luxonis_dataset)
    # all_payloads = _get_sample_payloads(luxonis_dataset)

    new_payloads = _filter_new_samples_by_id(vectordb_api, all_payloads)

    new_img_paths = [payload["image_path"] for payload in new_payloads]
    new_embeddings, succ_ix = extract_embeddings(
        new_img_paths,
        ort_session,
        luxonis_dataset.fs,
        transform,
        output_layer_name,
        emb_batch_size,
    )
    new_payloads = [new_payloads[ix] for ix in succ_ix]

    _batch_upsert(
        vectordb_api, new_embeddings, new_payloads, vectordb_batch_size
    )

    # make a instance_id : embedding dictionary
    instance_id_to_embedding = {
        payload["instance_id"]: embedding
        for payload, embedding in zip(new_payloads, new_embeddings)
    }
    print("Embeddings generation and insertion completed!")

    # returns only the new embeddings
    return instance_id_to_embedding



from pathlib import Path, PurePosixPath
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import typeguard
from pydantic import BaseModel
from typing_extensions import TypeAlias, TypeGuard

# When used without installed dependencies
if TYPE_CHECKING:  # pragma: no cover
    import numpy as np


PathType: TypeAlias = Union[str, Path]
"""A string or a `pathlib.Path` object."""

PosixPathType: TypeAlias = Union[str, PurePosixPath]
"""A string or a `pathlib.PurePosixPath` object."""


TaskType: TypeAlias = Literal[
    "classification",
    "boundingbox",
    "segmentation",
    "instance_segmentation",
    "keypoints",
    "array",
]


Labels: TypeAlias = Dict[str, "np.ndarray"]
"""Dictionary mappping task names to the annotations as C{np.ndarray}"""


LoaderOutput: TypeAlias = Tuple["np.ndarray", Labels]
"""C{LoaderOutput} is a tuple of an image as a C{np.ndarray} and a
dictionary of task group names and their annotations as
L{Annotations}."""

RGB: TypeAlias = Tuple[int, int, int]

HSV: TypeAlias = Tuple[float, float, float]

Color: TypeAlias = Union[str, int, RGB]
"""Color type alias.

Can be either a string (e.g. "red", "#FF5512"),  a tuple of RGB values,
or a single value (in which case it is interpreted as a grayscale
value).
"""

PrimitiveType: TypeAlias = Union[str, int, float, bool, None]
"""Primitive types in Python."""

# To avoid infinite recursion
if TYPE_CHECKING:  # pragma: no cover
    ParamValue: TypeAlias = Union[
        Dict[PrimitiveType, "ParamValue"],
        List["ParamValue"],
        PrimitiveType,
    ]
else:
    ParamValue: TypeAlias = Any

Params: TypeAlias = Dict[str, ParamValue]
"""A keyword dictionary of additional parameters.

Usually loaded from a YAML file.
"""

Kwargs: TypeAlias = Dict[str, Any]
"""A keyword dictionary of arbitrary parameters."""


class ConfigItem(BaseModel):
    """Configuration schema for dynamic object instantiation. Typically
    used to instantiate objects stored in registries.

    A dictionary with a name and a dictionary of parameters.

    @type name: str
    @ivar name: The name of the object this configuration applies to.
        Required.
    @type params: Dict[str, JsonDict]
    @ivar params: Additional parameters for instantiating the object.
        Not required.
    """

    name: str

    params: Params = {}


T = TypeVar("T")


def check_type(value: Any, type_: Type[T]) -> TypeGuard[T]:
    """Checks if the value has the correct type.

    @type value: Any
    @param value: The value to check.
    @type type_: Type[K]
    @param type_: The type to check against.
    @rtype: bool
    @return: C{True} if the value has the correct type, C{False}
        otherwise.
    """
    try:
        typeguard.check_type(value, type_)
    except (typeguard.TypeCheckError, TypeError):
        return False
    return True


def all_not_none(values: Iterable[Any]) -> bool:
    """Checks if none of the values in the iterable is C{None}

    @type values: Iterable[Any]
    @param values: An iterable of values
    @rtype: bool
    @return: C{True} if all values are not C{None}, C{False} otherwise
    """
    return all(v is not None for v in values)


def any_not_none(values: Iterable[Any]) -> bool:
    """Checks if at least one value in the iterable is not C{None}

    @type values: Iterable[Any]
    @param values: An iterable of values
    @rtype: bool
    @return: C{True} if at least one value is not C{None}, C{False}
        otherwise
    """
    return any(v is not None for v in values)



import numpy as np

from luxonis_ml.data.utils import rgb_to_bool_masks


def test_rgb_to_bool_masks():
    segmentation_mask = np.array(
        [
            [[0, 0, 0], [255, 0, 0], [0, 255, 0]],
            [[0, 0, 0], [0, 255, 0], [0, 0, 255]],
        ],
        dtype=np.uint8,
    )

    class_colors = {
        "red": (255, 0, 0),
        "green": (0, 255, 0),
        "blue": (0, 0, 255),
    }

    expected_results = {
        "background": np.array([[True, False, False], [True, False, False]]),
        "red": np.array([[False, True, False], [False, False, False]]),
        "green": np.array([[False, False, True], [False, True, False]]),
        "blue": np.array([[False, False, False], [False, False, True]]),
    }

    for class_name, mask in rgb_to_bool_masks(
        segmentation_mask, class_colors, add_background_class=True
    ):
        assert class_name in expected_results
        assert np.array_equal(mask, expected_results[class_name])



from typing import Final

import numpy as np

from luxonis_ml.data.augmentations.custom.letterbox_resize import (
    LetterboxResize,
)

WIDTH: Final[int] = 640
HEIGHT: Final[int] = 480


def test_letterbox():
    img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
    letterbox = LetterboxResize(HEIGHT, WIDTH, p=1.0)
    x = letterbox(image=img, labels={})
    assert x["image"].shape == (HEIGHT, WIDTH, 3)



import json
import random
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union, cast

import cv2
import numpy as np
import polars as pl
import yaml
from loguru import logger
from typing_extensions import override

from luxonis_ml.data.augmentations import (
    AUGMENTATION_ENGINES,
    AugmentationEngine,
)
from luxonis_ml.data.datasets import (
    Annotation,
    Category,
    LuxonisDataset,
    UpdateMode,
    load_annotation,
)
from luxonis_ml.data.loaders.base_loader import BaseLoader
from luxonis_ml.data.utils import (
    get_task_name,
    get_task_type,
    split_task,
    task_type_iterator,
)
from luxonis_ml.data.utils.task_utils import task_is_metadata
from luxonis_ml.typing import Labels, LoaderOutput, Params, PathType


class LuxonisLoader(BaseLoader):
    def __init__(
        self,
        dataset: LuxonisDataset,
        view: Union[str, List[str]] = "train",
        augmentation_engine: Union[
            Literal["albumentations"], str  # noqa: PYI051
        ] = "albumentations",
        augmentation_config: Optional[Union[List[Params], PathType]] = None,
        height: Optional[int] = None,
        width: Optional[int] = None,
        keep_aspect_ratio: bool = True,
        exclude_empty_annotations: bool = False,
        color_space: Literal["RGB", "BGR"] = "RGB",
        *,
        keep_categorical_as_strings: bool = False,
        update_mode: Union[
            UpdateMode, Literal["all", "missing"]
        ] = UpdateMode.ALL,
        filter_task_names: Optional[List[str]] = None,
    ) -> None:
        """A loader class used for loading data from L{LuxonisDataset}.

        @type dataset: LuxonisDataset
        @param dataset: Instance of C{LuxonisDataset} to use.
        @type view: Union[str, List[str]]
        @param view: What splits to use. Can be either a single split or
            a list of splits. Defaults to C{"train"}.
        @type augmentation_engine: Union[Literal["albumentations"], str]
        @param augmentation_engine: The augmentation engine to use.
            Defaults to C{"albumentations"}.
        @type augmentation_config: Optional[Union[List[Params],
            PathType]]
        @param augmentation_config: The configuration for the
            augmentations. This can be either a list of C{Dict[str, JsonValue]} or
            a path to a configuration file.
            The config member is a dictionary with two keys: C{name} and
            C{params}. C{name} is the name of the augmentation to
            instantiate and C{params} is an optional dictionary
            of parameters to pass to the augmentation.

            Example::

                [
                    {"name": "HorizontalFlip", "params": {"p": 0.5}},
                    {"name": "RandomBrightnessContrast", "params": {"p": 0.1}},
                    {"name": "Defocus"}
                ]

        @type height: Optional[int]
        @param height: The height of the output images. Defaults to
            C{None}.
        @type width: Optional[int]
        @param width: The width of the output images. Defaults to
            C{None}.
        @type keep_aspect_ratio: bool
        @param keep_aspect_ratio: Whether to keep the aspect ratio of the
            images. Defaults to C{True}.
        @type color_space: Literal["RGB", "BGR"]
        @param color_space: The color space of the output images. Defaults
            to C{"RGB"}.
        @type exclude_empty_annotations: bool
        @param exclude_empty_annotations: Whether to exclude
            empty annotations from the final label dictionary.
            Defaults to C{False} (i.e. include empty annotations).

        @type keep_categorical_as_strings: bool
        @param keep_categorical_as_strings: Whether to keep categorical
            metadata labels as strings.
            Defaults to C{False} (i.e. convert categorical labels to integers).

        @type update_mode: UpdateMode
        @param update_mode: Enum that determines the sync mode for media files of the remote dataset (annotations and metadata are always overwritten):
            - UpdateMode.MISSING: Downloads only the missing media files for the dataset.
            - UpdateMode.ALL: Always downloads and overwrites all media files in the local dataset.

        @type filter_task_names: Optional[List[str]]
        @param filter_task_names: List of task names to filter the dataset by.
            If C{None}, all task names are included. Defaults to C{None}.
            This is useful for filtering out tasks that are not needed for a specific use case.
        """

        self.exclude_empty_annotations = exclude_empty_annotations
        self.color_space: Literal["RGB", "BGR"] = color_space
        self.height = height
        self.width = width

        self.dataset = dataset
        self.sync_mode = self.dataset.is_remote
        self.keep_categorical_as_strings = keep_categorical_as_strings
        self.filter_task_names = filter_task_names

        if self.sync_mode:
            self.dataset.pull_from_cloud(update_mode=UpdateMode(update_mode))

        if isinstance(view, str):
            view = [view]
        self.view = view

        self.df = self.dataset._load_df_offline(raise_when_empty=True)
        self.classes = self.dataset.get_classes()

        if self.filter_task_names is not None:
            self.df = self.df.filter(
                pl.col("task_name").is_in(self.filter_task_names)
            )
            self.classes = {
                task_name: self.classes[task_name]
                for task_name in self.filter_task_names
                if task_name in self.classes
            }

        if not self.dataset.is_remote:
            file_index = self.dataset._get_index()
            if file_index is None:  # pragma: no cover
                raise FileNotFoundError("Cannot find file index")
            file_index = file_index.filter(
                pl.col("uuid").is_in(self.df["uuid"])
            )
            self.df = self.df.join(file_index, on="uuid").drop("file_right")

        self.instances: List[str] = []
        splits_path = self.dataset.metadata_path / "splits.json"
        if not splits_path.exists():
            raise RuntimeError(
                "Cannot find splits! Ensure you call dataset.make_splits()"
            )
        with open(splits_path) as file:
            splits = json.load(file)

        for view in self.view:
            self.instances.extend(splits[view])

        self.instances = [
            uuid
            for uuid in self.instances
            if uuid in self.df["uuid"].to_list()
        ]

        self.idx_to_df_row: List[List[int]] = []
        for uuid in self.instances:
            boolean_mask = self.df["uuid"] == uuid
            row_indexes = boolean_mask.arg_true().to_list()
            self.idx_to_df_row.append(row_indexes)

        self.tasks_without_background = set()

        self._precompute_image_paths()

        _, test_labels = self._load_data(0)
        for task, seg_masks in task_type_iterator(test_labels, "segmentation"):
            task_name = get_task_name(task)
            if seg_masks.shape[0] > 1 and (
                "background" not in self.classes[task_name]
                or self.classes[task_name]["background"] != 0
            ):
                unassigned_pixels = np.sum(seg_masks, axis=0) == 0

                if np.any(unassigned_pixels):
                    logger.warning(
                        "Found unassigned pixels in segmentation masks. "
                        "Assigning them to `background` class (class index 0). "
                        "If this is not desired then make sure all pixels are "
                        "assigned to one class or rename your background class."
                    )
                    self.tasks_without_background.add(task)
                    if "background" not in self.classes[task_name]:
                        self.classes[task_name] = {
                            "background": 0,
                            **{
                                class_name: i + 1
                                for class_name, i in self.classes[
                                    task_name
                                ].items()
                            },
                        }

        self.augmentations = self._init_augmentations(
            augmentation_engine,
            augmentation_config or [],
            height,
            width,
            keep_aspect_ratio,
        )

    @override
    def __len__(self) -> int:
        """Returns length of the dataset.

        @rtype: int
        @return: Length of the loader.
        """
        return len(self.instances)

    @override
    def __getitem__(self, idx: int) -> LoaderOutput:
        """Function to load a sample consisting of an image and its
        annotations.

        @type idx: int
        @param idx: The integer index of the sample to retrieve.
        @rtype: L{LuxonisLoaderOutput}
        @return: The loader ouput consisting of the image and a
            dictionary defining its annotations.
        """

        if self.augmentations is None:
            img, labels = self._load_data(idx)
        else:
            img, labels = self._load_with_augmentations(idx)

        if self.color_space == "BGR":
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        if self.exclude_empty_annotations:
            return img, labels

        return self._add_empty_annotations(img, labels)

    def _add_empty_annotations(
        self, img: np.ndarray, labels: Labels
    ) -> LoaderOutput:
        for task_name, task_types in self.dataset.get_tasks().items():
            if (
                self.filter_task_names is not None
                and task_name not in self.filter_task_names
            ):
                continue
            for task_type in task_types:
                task = f"{task_name}/{task_type}"
                if task not in labels:
                    if task_type == "boundingbox":
                        labels[task] = np.zeros((0, 5))
                    elif task_type == "keypoints":
                        n_keypoints = self.dataset.get_n_keypoints()[task_name]
                        labels[task] = np.zeros((0, n_keypoints * 3))
                    elif task_type == "instance_segmentation":
                        labels[task] = np.zeros(
                            (0, img.shape[0], img.shape[1])
                        )
                    elif task_type == "segmentation":
                        labels[task] = np.zeros(
                            (
                                len(self.classes[task_name]),
                                img.shape[0],
                                img.shape[1],
                            )
                        )
                    elif task_type == "classification" or task_is_metadata(
                        task
                    ):
                        labels[task] = np.zeros(
                            (len(self.classes[task_name]),)
                        )

        return img, labels

    def _load_data(self, idx: int) -> Tuple[np.ndarray, Labels]:
        """Loads image and its annotations based on index.

        @type idx: int
        @param idx: Index of the image
        @rtype: Tuple[np.ndarray, Labels]
        @return: Image as C{np.ndarray} in RGB format and a dictionary
            with all the present annotations
        """

        if not self.idx_to_df_row:
            raise ValueError(
                f"No data found in dataset '{self.dataset.identifier}' "
                f"for {self.view} views"
            )

        ann_indices = self.idx_to_df_row[idx]

        ann_rows = [self.df.row(row) for row in ann_indices]

        img_path = self.idx_to_img_path[idx]

        img = cv2.cvtColor(cv2.imread(str(img_path)), cv2.COLOR_BGR2RGB)

        labels_by_task: Dict[str, List[Annotation]] = defaultdict(list)
        class_ids_by_task: Dict[str, List[int]] = defaultdict(list)
        instance_ids_by_task: Dict[str, List[int]] = defaultdict(list)
        metadata_by_task: Dict[str, List[Union[str, int, float, Category]]] = (
            defaultdict(list)
        )

        for annotation_data in ann_rows:
            task_name: str = annotation_data[2]
            class_name: Optional[str] = annotation_data[3]
            instance_id: int = annotation_data[4]
            task_type: str = annotation_data[5]
            ann_str: Optional[str] = annotation_data[6]

            if ann_str is None:
                continue

            data = json.loads(ann_str)
            full_task_name = f"{task_name}/{task_type}"
            task_type = get_task_type(full_task_name)
            if task_type == "array" and self.dataset.is_remote:
                data["path"] = self.dataset.arrays_path / data["path"]

            if task_type.startswith("metadata/"):
                metadata_by_task[full_task_name].append(data)
            else:  # pragma: no cover
                # Conversion from LDF v1.0
                if "points" in data and "width" not in data:
                    data["width"] = img.shape[1]
                    data["height"] = img.shape[0]
                    data["points"] = [tuple(p) for p in data["points"]]

                annotation = load_annotation(task_type, data)
                labels_by_task[full_task_name].append(annotation)
                if class_name is not None:
                    class_ids_by_task[full_task_name].append(
                        self.classes[task_name][class_name]
                    )
                else:
                    class_ids_by_task[full_task_name].append(0)
                instance_ids_by_task[full_task_name].append(instance_id)

        labels: Labels = {}
        encodings = self.dataset.get_categorical_encodings()
        for task, metadata in metadata_by_task.items():
            if not self.keep_categorical_as_strings and task in encodings:
                metadata = [encodings[task][m] for m in metadata]  # type: ignore
            labels[task] = np.array(metadata)

        for task, anns in labels_by_task.items():
            assert anns, f"No annotations found for task {task_name}"
            instance_ids = instance_ids_by_task[task]

            anns = [
                ann
                for _, ann in sorted(
                    zip(instance_ids, anns), key=lambda x: x[0]
                )
            ]

            task_name, task_type = split_task(task)
            array = anns[0].combine_to_numpy(
                anns,
                class_ids_by_task[task],
                len(self.classes[task_name]),
            )
            if task in self.tasks_without_background:
                unassigned_pixels = ~np.any(array, axis=0)
                background_idx = self.classes[task_name]["background"]
                array[background_idx, unassigned_pixels] = 1

            labels[task] = array

        return img, labels

    def _load_with_augmentations(self, idx: int) -> LoaderOutput:
        indices = [idx]
        assert self.augmentations is not None
        if self.augmentations.batch_size > 1:
            if self.augmentations.batch_size > len(self):
                warnings.warn(
                    f"Augmentations batch_size ({self.augmentations.batch_size}) "
                    f"is larger than dataset size ({len(self)}). "
                    "Samples will include repetitions.",
                    stacklevel=2,
                )
                other_indices = [i for i in range(len(self)) if i != idx]
                picked_indices = random.choices(
                    other_indices, k=self.augmentations.batch_size - 1
                )
            else:
                picked_indices = set()
                max_val = len(self)
                while len(picked_indices) < self.augmentations.batch_size - 1:
                    rand_idx = random.randint(0, max_val - 1)
                    if rand_idx != idx and rand_idx not in picked_indices:
                        picked_indices.add(rand_idx)
                picked_indices = list(picked_indices)

            indices.extend(picked_indices)

        loaded_anns = [self._load_data(i) for i in indices]
        return self.augmentations.apply(loaded_anns)

    def _init_augmentations(
        self,
        augmentation_engine: Union[
            Literal["albumentations"],  # noqa: PYI051
            str,
        ],
        augmentation_config: Union[List[Params], PathType],
        height: Optional[int],
        width: Optional[int],
        keep_aspect_ratio: bool,
    ) -> Optional[AugmentationEngine]:
        if isinstance(augmentation_config, (Path, str)):
            with open(augmentation_config) as file:
                augmentation_config = cast(
                    List[Params], yaml.safe_load(file) or []
                )
        if augmentation_config and (width is None or height is None):
            raise ValueError(
                "Height and width must be provided when using augmentations"
            )

        if height is None or width is None:
            return None

        dataset_tasks = self.dataset.get_tasks()

        targets = {
            f"{task_name}/{task_type}": task_type
            for task_name, task_types in dataset_tasks.items()
            if self.filter_task_names is None
            or task_name in self.filter_task_names
            for task_type in task_types
        }

        n_classes = {
            f"{task_name}/{task_type}": self.dataset.get_n_classes()[task_name]
            for task_name, task_types in dataset_tasks.items()
            if self.filter_task_names is None
            or task_name in self.filter_task_names
            for task_type in task_types
        }

        return AUGMENTATION_ENGINES.get(augmentation_engine)(
            height=height,
            width=width,
            config=augmentation_config,
            targets=targets,
            n_classes=n_classes,
            keep_aspect_ratio=keep_aspect_ratio,
            is_validation_pipeline="train" not in self.view,
        )

    def _precompute_image_paths(self) -> None:
        self.idx_to_img_path = {}

        for idx, ann_indices in enumerate(self.idx_to_df_row):
            ann_indices = self.idx_to_df_row[idx]

            ann_rows = [self.df.row(row) for row in ann_indices]
            img_path = ann_rows[0][0]
            if not Path(img_path).exists():
                uuid = ann_rows[0][7]
                file_extension = ann_rows[0][0].rsplit(".", 1)[-1]
                img_path = self.dataset.media_path / f"{uuid}.{file_extension}"
                if not img_path.exists():
                    raise FileNotFoundError(
                        f"Cannot find image for uuid {uuid}"
                    )

            self.idx_to_img_path[idx] = img_path



import math
import random
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from albumentations.core.bbox_utils import denormalize_bboxes, normalize_bboxes
from typing_extensions import override

from luxonis_ml.data.augmentations.batch_transform import BatchTransform


class Mosaic4(BatchTransform):
    def __init__(
        self,
        out_height: int,
        out_width: int,
        value: Optional[Union[int, float, List[int], List[float]]] = None,
        mask_value: Optional[Union[int, float, List[int], List[float]]] = None,
        p: float = 0.5,
    ):
        """Mosaic augmentation arranges selected four images into single
        image in a 2x2 grid layout. This is done in deterministic way
        meaning first image in the batch will always be in top left. The
        input images should have the same number of channels but can
        have different widths and heights. The output is cropped around
        the intersection point of the four images with the size
        (out_with x out_height). If the mosaic image is smaller than
        width x height, the gap is filled by the C{fill_value}.

        @type out_height: int
        @param out_height: Output image height. The mosaic image is
            cropped by this height around the mosaic center. If the size
            of the mosaic image is smaller than this value the gap is
            filled by the C{value}.
        @type out_width: int
        @param out_width: Output image width. The mosaic image is
            cropped by this height around the mosaic center. If the size
            of the mosaic image is smaller than this value the gap is
            filled by the C{value}.
        @type value: Optional[Union[int, float, List[int], List[float]]]
        @param value: Padding value. Defaults to C{None}.
        @type mask_value: Optional[Union[int, float, List[int],
            List[float]]]
        @param mask_value: Padding value for masks. Defaults to C{None}.
        @type p: float
        @param p: Probability of applying the transform. Defaults to
            C{0.5}.
        """

        super().__init__(batch_size=4, p=p)

        if out_height <= 0:
            raise ValueError(
                f"`out_height` must be larger than 0, got {out_height}"
            )
        if out_width <= 0:
            raise ValueError(
                f"`out_width` must be larger than 0, got {out_width}"
            )

        self.out_height = out_height
        self.out_width = out_width
        self.value = value
        self.mask_value = mask_value

    def generate_random_crop_center(self) -> Tuple[int, int]:
        """Generate a random crop center within the bounds of the mosaic
        image size."""
        crop_x = random.randint(0, max(0, self.out_width))
        crop_y = random.randint(0, max(0, self.out_height))
        return crop_x, crop_y

    @override
    def apply(
        self, image_batch: List[np.ndarray], x_crop: int, y_crop: int, **_
    ) -> np.ndarray:
        """Applies the transformation to a batch of images.

        @type image_batch: List[np.ndarray]
        @param image_batch: Batch of input images to which the
            transformation is applied.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: np.ndarray
        @return: Transformed images.
        """
        return apply_mosaic4_to_images(
            image_batch,
            self.out_height,
            self.out_width,
            x_crop,
            y_crop,
            self.value,
        )

    @override
    def apply_to_mask(
        self,
        mask_batch: List[np.ndarray],
        x_crop: int,
        y_crop: int,
        cols: int,
        rows: int,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of masks.

        @type mask_batch: List[np.ndarray]
        @param mask_batch: Batch of input masks to which the
            transformation is applied.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: np.ndarray
        @return: Transformed masks.
        """
        for i in range(len(mask_batch)):
            mask = mask_batch[i]
            if mask.size == 0:
                if len(mask.shape) == 2:
                    mask_batch[i] = np.zeros((rows, cols), dtype=mask.dtype)
                else:
                    mask_batch[i] = np.zeros(
                        (rows, cols, mask.shape[-1]), dtype=mask.dtype
                    )
        return apply_mosaic4_to_images(
            mask_batch,
            self.out_height,
            self.out_width,
            x_crop,
            y_crop,
            self.mask_value,
        )

    @override
    def apply_to_instance_mask(
        self, masks_batch: List[np.ndarray], x_crop: int, y_crop: int, **_
    ) -> np.ndarray:
        """Applies the transformation to a batch of instance masks.

        @type mask_batch: List[np.ndarray]
        @param mask_batch: Batch of input masks to which the
            transformation is applied.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: np.ndarray
        @return: Transformed masks.
        """
        return apply_mosaic4_to_instance_masks(
            masks_batch,
            self.out_height,
            self.out_width,
            x_crop,
            y_crop,
            self.value,
        )

    @override
    def apply_to_bboxes(
        self,
        bboxes_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        x_crop: int,
        y_crop: int,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of bboxes.

        @type bboxes_batch: List[np.ndarray]
        @param bboxes_batch: Batch of input bboxes to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[np.ndarray]
        @return: List of transformed bboxes.
        """
        new_bboxes = []
        for i, (bboxes, (rows, cols)) in enumerate(
            zip(bboxes_batch, image_shapes)
        ):
            if bboxes.size == 0:  # pragma: no cover
                bboxes = np.zeros((0, 6), dtype=bboxes.dtype)

            bbox = apply_mosaic4_to_bboxes(
                bboxes,
                rows,
                cols,
                i,
                self.out_height,
                self.out_width,
                x_crop,
                y_crop,
            )
            new_bboxes.append(bbox)

        return np.concatenate(new_bboxes, axis=0)

    @override
    def apply_to_keypoints(
        self,
        keypoints_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        x_crop: int,
        y_crop: int,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of keypoints.

        @type keypoints_batch: List[KeypointType]
        @param keypoints_batch: Batch of input keypoints to which the
            transformation is applied.
        @type indices: List[Tuple[int, int]]
        @param indices: Indices of images in the batch.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type params: Any
        @param params: Additional parameters for the transformation.
        @type x_crop: int
        @param x_crop: x-coordinate of the croping start point
        @type y_crop: int
        @param y_crop: y-coordinate of the croping start point
        @rtype: List[KeypointType]
        @return: List of transformed keypoints.
        """
        new_keypoints = []
        for i, (keypoints, (rows, cols)) in enumerate(
            zip(keypoints_batch, image_shapes)
        ):
            if keypoints.size == 0:
                keypoints = np.zeros((0, 5), dtype=keypoints.dtype)

            new_keypoint = apply_mosaic4_to_keypoints(
                keypoints,
                rows,
                cols,
                i,
                self.out_height,
                self.out_width,
                x_crop,
                y_crop,
            )
            new_keypoints.append(new_keypoint)
        return np.concatenate(new_keypoints, axis=0)

    @override
    def get_params_dependent_on_data(
        self, params: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Get parameters dependent on the targets.

        @type params: Dict[str, Any]
        @param params: Dictionary containing parameters.
        @rtype: Dict[str, Any]
        @return: Dictionary containing parameters dependent on the
            targets.
        """
        additional_params = super().get_params_dependent_on_data(params, data)
        image_batch = data["image"]
        image_shapes = [tuple(image.shape[:2]) for image in image_batch]
        x_crop, y_crop = self.generate_random_crop_center()
        additional_params.update(
            {"image_shapes": image_shapes, "x_crop": x_crop, "y_crop": y_crop}
        )
        return additional_params


def compute_mosaic4_corners(
    quadrant: int,
    out_height: int,
    out_width: int,
    in_height: int,
    in_width: int,
) -> Tuple[Tuple[int, int, int, int], Tuple[int, int, int, int]]:
    if quadrant == 0:
        x1a, y1a, x2a, y2a = (
            max(out_width - in_width, 0),
            max(out_height - in_height, 0),
            out_width,
            out_height,
        )
        x1b, y1b, x2b, y2b = (
            in_width - (x2a - x1a),
            in_height - (y2a - y1a),
            in_width,
            in_height,
        )

    elif quadrant == 1:
        x1a, y1a, x2a, y2a = (
            out_width,
            max(out_height - in_height, 0),
            min(out_width + in_width, out_width * 2),
            out_height,
        )
        x1b, y1b, x2b, y2b = (
            0,
            in_height - (y2a - y1a),
            min(in_width, x2a - x1a),
            in_height,
        )
    elif quadrant == 2:
        x1a, y1a, x2a, y2a = (
            max(out_width - in_width, 0),
            out_height,
            out_width,
            min(out_height * 2, out_height + in_height),
        )
        x1b, y1b, x2b, y2b = (
            in_width - (x2a - x1a),
            0,
            in_width,
            min(y2a - y1a, in_height),
        )
    else:
        x1a, y1a, x2a, y2a = (
            out_width,
            out_height,
            min(out_width + in_width, out_width * 2),
            min(out_height * 2, out_height + in_height),
        )
        x1b, y1b, x2b, y2b = (
            0,
            0,
            min(in_width, x2a - x1a),
            min(y2a - y1a, in_height),
        )
    return (x1a, y1a, x2a, y2a), (x1b, y1b, x2b, y2b)


def apply_mosaic4_to_instance_masks(
    masks_batch: List[np.ndarray],
    out_height: int,
    out_width: int,
    x_crop: int,
    y_crop: int,
    value: Optional[Union[int, float, List[int], List[float]]] = None,
) -> np.ndarray:
    out_masks = []
    out_shape = [out_height * 2, out_width * 2]

    if not any(m.size for m in masks_batch):
        return np.zeros((out_height, out_width, 0), dtype=np.uint8)

    for quadrant, masks in enumerate(masks_batch):
        if masks.size == 0:
            continue

        for i in range(masks.shape[-1]):
            mask = masks[..., i]
            imgsz = max(out_height, out_width)
            h, w = mask.shape
            r = imgsz / max(h, w)
            if r != 1:
                w, h = (
                    min(math.ceil(w * r), imgsz),
                    min(math.ceil(h * r), imgsz),
                )
                mask = cv2.resize(
                    mask, (w, h), interpolation=cv2.INTER_NEAREST
                )

            combined_mask = np.full(
                out_shape, value if value is not None else 0, dtype=masks.dtype
            )
            (x1a, y1a, x2a, y2a), (x1b, y1b, x2b, y2b) = (
                compute_mosaic4_corners(quadrant, out_height, out_width, h, w)
            )

            combined_region = combined_mask[y1a:y2a, x1a:x2a]
            mask_region = mask[y1b:y2b, x1b:x2b]

            combined_height, combined_width = combined_region.shape[:2]
            img_h, img_w = mask_region.shape[:2]

            min_h = min(combined_height, img_h)
            min_w = min(combined_width, img_w)

            combined_mask[y1a : y1a + min_h, x1a : x1a + min_w] = mask[
                y1b : y1b + min_h, x1b : x1b + min_w
            ]

            combined_mask = combined_mask[
                y_crop : y_crop + out_height, x_crop : x_crop + out_width
            ]
            out_masks.append(combined_mask)

    if not out_masks:
        return np.zeros((out_height, out_width, 0), dtype=np.uint8)

    return np.stack(out_masks, axis=-1)


def apply_mosaic4_to_images(
    image_batch: List[np.ndarray],
    out_height: int,
    out_width: int,
    x_crop: int,
    y_crop: int,
    padding: Optional[Union[int, float, List[int], List[float]]] = None,
) -> np.ndarray:
    """Arrange the images in a 2x2 grid layout.

    The input images should have the same number of channels but can
    have different widths and heights. The gaps are filled by the
    padding value.
    """

    if len(image_batch[0].shape) == 2:
        out_shape = [out_height * 2, out_width * 2]
    else:
        out_shape = [out_height * 2, out_width * 2, image_batch[0].shape[2]]

    combined_image = np.full(
        out_shape,
        padding if padding is not None else 0,
        dtype=image_batch[0].dtype,
    )

    for quadrant, img in enumerate(image_batch):
        h, w = img.shape[:2]
        imgsz = max(out_height, out_width)
        r = imgsz / max(h, w)
        if r != 1:
            w, h = (min(math.ceil(w * r), imgsz), min(math.ceil(h * r), imgsz))
            img = cv2.resize(img, (w, h), interpolation=cv2.INTER_LINEAR)
            if img.ndim == 2:
                img = img.reshape(h, w, 1)

        (x1a, y1a, x2a, y2a), (x1b, y1b, x2b, y2b) = compute_mosaic4_corners(
            quadrant, out_height, out_width, h, w
        )

        combined_region = combined_image[y1a:y2a, x1a:x2a]
        img_region = img[y1b:y2b, x1b:x2b]

        combined_height, combined_width = combined_region.shape[:2]
        img_h, img_w = img_region.shape[:2]

        min_h = min(combined_height, img_h)
        min_w = min(combined_width, img_w)

        combined_image[y1a : y1a + min_h, x1a : x1a + min_w] = img[
            y1b : y1b + min_h, x1b : x1b + min_w
        ]

    return combined_image[
        y_crop : y_crop + out_height, x_crop : x_crop + out_width
    ]


def apply_mosaic4_to_bboxes(
    bbox: np.ndarray,
    in_height: int,
    in_width: int,
    position_index: int,
    out_height: int,
    out_width: int,
    x_crop: int,
    y_crop: int,
) -> np.ndarray:
    """Adjust bounding box coordinates to account for mosaic grid
    position.

    This function modifies bounding boxes according to their placement
    in a 2x2 grid mosaic, shifting their coordinates based on the tile's
    relative position within the mosaic.
    """

    bbox = denormalize_bboxes(bbox, (in_height, in_width))

    imgsz = max(out_height, out_width)
    r = imgsz / max(in_height, in_width)
    if r != 1:
        in_width, in_height = (
            min(math.ceil(in_width * r), imgsz),
            min(math.ceil(in_height * r), imgsz),
        )
        bbox[:, :4] = bbox[:, :4] * r

    if position_index == 0:
        shift_x = out_width - in_width
        shift_y = out_height - in_height
    elif position_index == 1:
        shift_x = out_width
        shift_y = out_height - in_height
    elif position_index == 2:
        shift_x = out_width - in_width
        shift_y = out_height
    elif position_index == 3:
        shift_x = out_width
        shift_y = out_height

    bbox[:, 0] += shift_x - x_crop
    bbox[:, 2] += shift_x - x_crop

    bbox[:, 1] += shift_y - y_crop
    bbox[:, 3] += shift_y - y_crop

    return normalize_bboxes(bbox, (out_height, out_width))


def apply_mosaic4_to_keypoints(
    keypoints: np.ndarray,
    in_height: int,
    in_width: int,
    position_index: int,
    out_height: int,
    out_width: int,
    x_crop: int,
    y_crop: int,
) -> np.ndarray:
    """Adjust keypoint coordinates based on mosaic grid position.

    This function adjusts the keypoint coordinates by placing them in
    one of the 2x2 mosaic grid cells, with shifts relative to the mosaic
    center.
    """
    imgsz = max(out_height, out_width)
    r = imgsz / max(in_height, in_width)
    if r != 1:
        in_width, in_height = (
            min(math.ceil(in_width * r), imgsz),
            min(math.ceil(in_height * r), imgsz),
        )
        keypoints[:, 0] = keypoints[:, 0] * r
        keypoints[:, 1] = keypoints[:, 1] * r

    if position_index == 0:
        shift_x = out_width - in_width
        shift_y = out_height - in_height
    elif position_index == 1:
        shift_x = out_width
        shift_y = out_height - in_height
    elif position_index == 2:
        shift_x = out_width - in_width
        shift_y = out_height
    elif position_index == 3:
        shift_x = out_width
        shift_y = out_height

    keypoints[:, 0] += shift_x - x_crop
    keypoints[:, 1] += shift_y - y_crop

    mask_invalid = (
        (keypoints[:, 0] < 0)
        | (keypoints[:, 0] > out_width)
        | (keypoints[:, 1] < 0)
        | (keypoints[:, 1] > out_height)
    )

    keypoints[:, -1] = np.where(mask_invalid, 0, keypoints[:, -1])
    return keypoints



from typing import Dict, List

import numpy as np
import pytest
from pytest_subtests.plugin import SubTests

from luxonis_ml.data.augmentations.custom.mixup import MixUp


@pytest.mark.parametrize("augmentation_data", [2], indirect=True)
@pytest.mark.parametrize("alpha", [0.5, 0.1])
def test_mixup(
    height: int,
    width: int,
    augmentation_data: Dict[str, List[np.ndarray]],
    subtests: SubTests,
    alpha: float,
):
    mixup = MixUp(p=1.0, alpha=alpha)
    transformed = mixup(**augmentation_data)  # type: ignore
    assert set(transformed.keys()) == set(augmentation_data.keys())
    for value in transformed.values():
        assert isinstance(value, np.ndarray)

    with subtests.test("image"):
        assert transformed["image"].shape == (height, width, 3)
        assert np.allclose(
            (
                augmentation_data["image"][0] * alpha
                + augmentation_data["image"][1] * (1 - alpha)
            ),
            transformed["image"],
        )

    with subtests.test("bboxes"):
        assert transformed["bboxes"].shape == (2, 6)
        assert np.allclose(
            np.concatenate(augmentation_data["bboxes"]), transformed["bboxes"]
        )

    with subtests.test("keypoints"):
        assert transformed["keypoints"].shape == (2, 5)
        assert np.allclose(
            np.concatenate(augmentation_data["keypoints"]),
            transformed["keypoints"],
        )

    with subtests.test("mask"):
        assert transformed["mask"].shape == (height, width, 1)
        assert np.allclose(
            augmentation_data["mask"][0].astype(bool)
            | augmentation_data["mask"][1].astype(bool),
            transformed["mask"].astype(bool),
        )


def test_invalid():
    with pytest.raises(ValueError, match="must be in range"):
        MixUp(alpha=(-1, 1))
    with pytest.raises(ValueError, match="must be in range"):
        MixUp(alpha=(1, -1))
    with pytest.raises(ValueError, match="ascending order"):
        MixUp(alpha=(0.8, 0.2))



from pathlib import Path
from typing import List

import pytest

from luxonis_ml.data import DatasetIterator, LuxonisLoader
from luxonis_ml.typing import Params
from tests.test_data.utils import create_dataset, create_image


@pytest.mark.parametrize("n_samples", range(20))
def test_empty(dataset_name: str, tempdir: Path, n_samples: int):
    config: List[Params] = [
        {
            "name": "Defocus",
            "params": {"p": 1.0},
        },
        {
            "name": "Mosaic4",
            "params": {"p": 1.0, "out_width": 640, "out_height": 640},
        },
    ]

    def generator(keep_samples: int) -> DatasetIterator:
        for i in range(20):
            img = create_image(i, tempdir)
            if i < keep_samples:
                yield {
                    "file": img,
                    "annotation": {
                        "class": ["dog", "cat"][i % 2],
                        "segmentation": {
                            "points": [(0.2, 0.2), (0.2, 0.4), (0.4, 0.4)],
                            "height": 256,
                            "width": 256,
                        },
                        "instance_segmentation": {
                            "height": 256,
                            "width": 256,
                            "points": [(0.25, 0.25), (0.3, 0.3), (0.4, 0.4)],
                        },
                        "boundingbox": {
                            "x": 0.2,
                            "y": 0.2,
                            "w": 0.3,
                            "h": 0.3,
                        },
                        "keypoints": {
                            "keypoints": [(0.25, 0.25, 2), (0.3, 0.3, 1)]
                        },
                        "metadata": {
                            "breed": "labrador",
                        },
                    },
                }
            else:
                yield {"file": img}

    dataset = create_dataset(
        dataset_name, generator(n_samples), splits={"train": 1.0}
    )
    loader = LuxonisLoader(
        dataset, augmentation_config=config, height=256, width=256
    )
    for _, labels in loader:
        if "/classification" not in labels:
            continue

        n_classes = dataset.get_n_classes()[""]
        if labels["/classification"].sum() == 0:
            assert labels["/classification"].shape == (n_classes,)
            assert labels["/boundingbox"].shape == (0, 5)
            assert labels["/keypoints"].shape == (0, 2 * 3)
            assert labels["/segmentation"].shape == (n_classes, 256, 256)
            assert labels["/segmentation"].sum() == 0
            assert labels["/instance_segmentation"].shape == (0, 256, 256)



from ..guard_extras import guard_missing_extra

with guard_missing_extra("embedd"):
    from .methods import (
        calculate_similarity_matrix,
        find_mismatches_centroids,
        find_mismatches_knn,
        find_representative_greedy,
        find_representative_kmedoids,
        find_similar,
        isolation_forest_OOD,
        leverage_OOD,
    )
    from .utils import (
        QdrantAPI,
        QdrantManager,
        VectorDBAPI,
        WeaviateAPI,
        extend_output_onnx,
        extract_embeddings,
        generate_embeddings,
        load_model_onnx,
        save_model_onnx,
    )

__all__ = [
    "QdrantAPI",
    "QdrantManager",
    "VectorDBAPI",
    "WeaviateAPI",
    "calculate_similarity_matrix",
    "extend_output_onnx",
    "extract_embeddings",
    "find_mismatches_centroids",
    "find_mismatches_knn",
    "find_representative_greedy",
    "find_representative_kmedoids",
    "find_similar",
    "generate_embeddings",
    "isolation_forest_OOD",
    "leverage_OOD",
    "load_model_onnx",
    "save_model_onnx",
]



from .embedding import extract_embeddings
from .ldf import generate_embeddings
from .model import extend_output_onnx, load_model_onnx, save_model_onnx
from .qdrant import QdrantAPI, QdrantManager
from .vectordb import VectorDBAPI
from .weaviate import WeaviateAPI

__all__ = [
    "QdrantAPI",
    "QdrantManager",
    "VectorDBAPI",
    "WeaviateAPI",
    "extend_output_onnx",
    "extract_embeddings",
    "generate_embeddings",
    "load_model_onnx",
    "save_model_onnx",
]



from .albumentations_engine import AlbumentationsEngine
from .base_engine import AUGMENTATION_ENGINES, AugmentationEngine
from .batch_compose import BatchCompose
from .custom import LetterboxResize, MixUp, Mosaic4

__all__ = [
    "AUGMENTATION_ENGINES",
    "AlbumentationsEngine",
    "AugmentationEngine",
    "BatchCompose",
    "LetterboxResize",
    "MixUp",
    "Mosaic4",
]



import json
from contextlib import suppress
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class NativeParser(BaseParser):
    """Parses directory with native LDF annotations.

    Expected format::

        dataset_dir/
        ├── train/
        │   └── annotations.json
        ├── valid/
        └── test/

    The annotations are stored in a single JSON file as a list of dictionaries
    in the same format as the output of the generator function used
    in L{BaseDataset.add} method.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        annotation_path = split_path / "annotations.json"
        if not annotation_path.exists():
            return None
        return {"annotation_path": annotation_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "val", "test"]:
            split_path = dataset_dir / split
            if NativeParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        added_train_imgs = self._parse_split(
            annotation_path=dataset_dir / "train" / "annotations.json",
        )
        added_val_imgs = self._parse_split(
            annotation_path=dataset_dir / "val" / "annotations.json",
        )
        added_test_imgs = self._parse_split(
            annotation_path=dataset_dir / "test" / "annotations.json",
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(self, annotation_path: Path) -> ParserOutput:
        """Parses annotations from LDF Format.

        @type annotation_path: C{Path}
        @param annotation_dir: Path to the JSON file with annotations.
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """

        data = json.loads(annotation_path.read_text())

        def generator() -> DatasetIterator:
            for record in data:
                with suppress(KeyError):
                    record["file"] = (
                        annotation_path.parent / record["file"]
                    ).absolute()
                for mask_type in ["segmentation", "instance_segmentation"]:
                    with suppress(KeyError):
                        mask = record["annotation"][mask_type]["mask"]
                        if isinstance(mask, (str, Path)):
                            record["annotation"][mask_type]["mask"] = (
                                annotation_path.parent / mask
                            ).absolute()
                yield record

        added_images = self._get_added_images(generator())

        return generator(), {}, added_images



from importlib.metadata import entry_points

from luxonis_ml.guard_extras import guard_missing_extra

with guard_missing_extra("data"):
    from .augmentations import AlbumentationsEngine
    from .datasets import (
        DATASETS_REGISTRY,
        BaseDataset,
        Category,
        DatasetIterator,
        LuxonisComponent,
        LuxonisDataset,
        LuxonisSource,
        Metadata,
        UpdateMode,
    )
    from .loaders import LOADERS_REGISTRY, BaseLoader, LuxonisLoader
    from .parsers import LuxonisParser
    from .utils.enums import BucketStorage, BucketType, ImageType, MediaType


def load_dataset_plugins() -> None:  # pragma: no cover
    """Registers any external dataset BaseDataset class plugins."""
    for entry_point in _get_entry_points_subset("dataset_plugins"):
        plugin_class = entry_point.load()
        DATASETS_REGISTRY.register(module=plugin_class)


def load_loader_plugins() -> None:  # pragma: no cover
    """Registers any external dataset BaseLoader class plugins."""
    for entry_point in _get_entry_points_subset("loader_plugins"):
        plugin_class = entry_point.load()
        DATASETS_REGISTRY.register(module=plugin_class)


def _get_entry_points_subset(key: str) -> list:
    """Returns subset of selected entry points.

    Safe function for older (py3.8) and newer py versions
    """
    entry_points_obj = entry_points()
    if isinstance(entry_points_obj, dict):
        # py3.8 specific
        selected_entry_points = entry_points_obj.get(key, [])
    else:
        selected_entry_points = entry_points_obj.select(group=key)
    return selected_entry_points


load_dataset_plugins()
load_loader_plugins()

__all__ = [
    "DATASETS_REGISTRY",
    "LOADERS_REGISTRY",
    "AlbumentationsEngine",
    "BaseDataset",
    "BaseLoader",
    "BucketStorage",
    "BucketType",
    "Category",
    "DatasetIterator",
    "ImageType",
    "LuxonisComponent",
    "LuxonisDataset",
    "LuxonisLoader",
    "LuxonisParser",
    "LuxonisSource",
    "MediaType",
    "Metadata",
    "UpdateMode",
]



import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from defusedxml.ElementTree import parse

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class VOCParser(BaseParser):
    """Parses directory with VOC annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img1.xml
        │   └── ...
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None

        images = BaseParser._list_images(split_path)
        labels = split_path.glob("*.xml")
        if not BaseParser._compare_stem_files(images, labels):
            return None
        return {"image_dir": split_path, "annotation_dir": split_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if VOCParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            annotation_dir=dataset_dir / "train",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            annotation_dir=dataset_dir / "valid",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test", annotation_dir=dataset_dir / "test"
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_dir: Path
    ) -> ParserOutput:
        """Parses annotations from VOC format to LDF. Annotations
        include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_dir: Path
        @param annotation_dir: Path to directory with C{.xml}
            annotations
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """

        images_annotations = []
        for anno_xml in annotation_dir.glob("*.xml"):
            annotation_data = parse(anno_xml)
            root = annotation_data.getroot()
            if root is None:
                raise ValueError(f"Could not parse {anno_xml}")

            path = image_dir.absolute().resolve() / self._xml_find(
                root, "filename"
            )
            if not path.exists():
                continue

            curr_annotations = {"path": path, "classes": [], "bboxes": []}
            size_item = root.find("size")
            assert size_item is not None
            height = float(self._xml_find(size_item, "height"))
            width = float(self._xml_find(size_item, "width"))

            for object_item in root.findall("object"):
                class_name = self._xml_find(object_item, "name")
                curr_annotations["classes"].append(class_name)

                bbox_info = object_item.find("bndbox")
                if bbox_info is not None:
                    bbox_xywh = np.array(
                        [
                            float(self._xml_find(bbox_info, "xmin")),
                            float(self._xml_find(bbox_info, "ymin")),
                            float(self._xml_find(bbox_info, "xmax"))
                            - float(self._xml_find(bbox_info, "xmin")),
                            float(self._xml_find(bbox_info, "ymax"))
                            - float(self._xml_find(bbox_info, "ymin")),
                        ]
                    )
                    bbox_xywh[::2] /= width
                    bbox_xywh[1::2] /= height
                    bbox_xywh = bbox_xywh.tolist()
                    curr_annotations["bboxes"].append((class_name, bbox_xywh))
            images_annotations.append(curr_annotations)

        def generator() -> DatasetIterator:
            for curr_annotations in images_annotations:
                path = str(curr_annotations["path"])
                for bbox_class, bbox in curr_annotations["bboxes"]:
                    yield {
                        "file": path,
                        "annotation": {
                            "class": bbox_class,
                            "boundingbox": {
                                "x": bbox[0],
                                "y": bbox[1],
                                "w": bbox[2],
                                "h": bbox[3],
                            },
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), {}, added_images

    @staticmethod
    def _xml_find(root: ET.Element, tag: str) -> str:
        item = root.find(tag)
        if item is not None and item.text is not None:
            return item.text
        raise ValueError(f"Could not find {tag} in {root}")



from pathlib import Path
from typing import List

import cv2
import numpy as np
import pydantic
import pytest
from pytest_subtests import SubTests

from luxonis_ml.data.datasets.annotation import (
    ArrayAnnotation,
    BBoxAnnotation,
    ClassificationAnnotation,
    DatasetRecord,
    Detection,
    InstanceSegmentationAnnotation,
    KeypointAnnotation,
    SegmentationAnnotation,
    check_valid_identifier,
    load_annotation,
)
from luxonis_ml.data.utils.parquet import ParquetRecord


def test_valid_identifier():
    check_valid_identifier("variable", label="")
    check_valid_identifier("variable_name", label="")
    check_valid_identifier("variable-name", label="")

    with pytest.raises(ValueError, match="can only contain alphanumeric"):
        check_valid_identifier("variable name", label="")

    with pytest.raises(ValueError, match="can only contain alphanumeric"):
        check_valid_identifier("?variable_name", label="")

    with pytest.raises(ValueError, match="can only contain alphanumeric"):
        check_valid_identifier("12variable_name", label="")

    with pytest.raises(ValueError, match="can only contain alphanumeric"):
        check_valid_identifier("variable/name", label="")


def test_load_annotation():
    assert load_annotation(
        "boundingbox", {"x": 0.1, "y": 0.2, "w": 0.3, "h": 0.4}
    ) == BBoxAnnotation(x=0.1, y=0.2, w=0.3, h=0.4)
    with pytest.raises(ValueError, match="Unknown label type"):
        load_annotation("invalid_name", {})


def test_dataset_record(tempdir: Path):
    def compare_parquet_rows(
        record: DatasetRecord, expected_rows: List[ParquetRecord]
    ) -> None:
        rows = list(record.to_parquet_rows())
        for row in rows:
            # for compatibility with Windows
            row["file"] = Path(row["file"])  # type: ignore
        assert rows == expected_rows

    cv2.imwrite(str(tempdir / "left.jpg"), np.zeros((100, 100, 3)))
    cv2.imwrite(str(tempdir / "right.jpg"), np.zeros((100, 100, 3)))
    record = DatasetRecord(file=tempdir / "left.jpg")  # type: ignore
    assert record.file == tempdir / "left.jpg"

    compare_parquet_rows(
        record,
        [
            {
                "file": tempdir / "left.jpg",  # type: ignore
                "source_name": "image",
                "task_name": "",
                "class_name": None,
                "instance_id": None,
                "task_type": None,
                "annotation": None,
            }
        ],
    )

    record = DatasetRecord(
        file=tempdir / "left.jpg",  # type: ignore
        annotation={
            "class": "person",
            "boundingbox": {"x": 0.1, "y": 0.2, "w": 0.3, "h": 0.4},
        },
    )
    compare_parquet_rows(
        record,
        [
            {
                "file": tempdir / "left.jpg",  # type: ignore
                "source_name": "image",
                "task_name": "",
                "class_name": "person",
                "instance_id": -1,
                "task_type": "boundingbox",
                "annotation": '{"x":0.1,"y":0.2,"w":0.3,"h":0.4}',
            },
            {
                "file": tempdir / "left.jpg",  # type: ignore
                "source_name": "image",
                "task_name": "",
                "class_name": "person",
                "instance_id": -1,
                "task_type": "classification",
                "annotation": "{}",
            },
        ],
    )

    record = DatasetRecord(
        files={
            "left": tempdir / "left.jpg",
            "right": tempdir / "right.jpg",
        }
    )
    with pytest.raises(ValueError, match="must have exactly one file"):
        _ = record.file


def test_bbox_annotation(subtests: SubTests):
    with subtests.test("simple"):
        bbox = BBoxAnnotation(x=0.1, y=0.2, w=0.3, h=0.4)
        assert bbox.x == 0.1
        assert bbox.y == 0.2
        assert bbox.w == 0.3
        assert bbox.h == 0.4

    with subtests.test("no_auto_clip"):
        base_dict = {"x": 0, "y": 0, "w": 0, "h": 0}
        for k in ["x", "y", "w", "h"]:
            for v in [-2.1, 2.3, -3.3, 3]:
                curr_dict = base_dict.copy()
                curr_dict[k] = v
                with pytest.raises(
                    ValueError, match="outside of automatic clipping"
                ):
                    BBoxAnnotation(**curr_dict)  # type: ignore

        bbox_ann = BBoxAnnotation(x=0.9, y=0, w=0.2, h=0)
        assert bbox_ann.x + bbox_ann.w <= 1
        bbox_ann = BBoxAnnotation(x=1.2, y=0, w=0.2, h=0)
        assert bbox_ann.x + bbox_ann.w <= 1
        bbox_ann = BBoxAnnotation(x=0, y=0.9, w=0, h=0.2)
        assert bbox_ann.y + bbox_ann.h <= 1
        bbox_ann = BBoxAnnotation(x=0, y=1.2, w=0, h=0.2)
        assert bbox_ann.y + bbox_ann.h <= 1

    with subtests.test("auto_clip"):
        base_dict = {"x": 0, "y": 0, "w": 0, "h": 0}
        for k in ["x", "y", "w", "h"]:
            for v in [-1.1, 1.3, -1.3, 2]:
                curr_dict = base_dict.copy()
                curr_dict[k] = v
                bbox_ann = BBoxAnnotation(**curr_dict)  # type: ignore
                assert 0 <= bbox_ann.x <= 1
                assert 0 <= bbox_ann.y <= 1
                assert 0 <= bbox_ann.w <= 1
                assert 0 <= bbox_ann.h <= 1

    with subtests.test("numpy"):
        bbox = BBoxAnnotation(x=0.1, y=0.2, w=0.3, h=0.4)
        assert np.allclose(
            bbox.to_numpy(class_id=4), np.array([4, 0.1, 0.2, 0.3, 0.4])
        )
        bboxes = [
            bbox,
            BBoxAnnotation(x=0.2, y=0.3, w=0.4, h=0.5),
            BBoxAnnotation(x=0.3, y=0.4, w=0.5, h=0.6),
        ]
        assert np.allclose(
            BBoxAnnotation.combine_to_numpy(bboxes, [1, 2, 3], ...),
            np.array(
                [
                    [1, 0.1, 0.2, 0.3, 0.4],
                    [2, 0.2, 0.3, 0.4, 0.5],
                    [3, 0.3, 0.4, 0.5, 0.6],
                ]
            ),
        )


def test_keypoints_annotation(subtests: SubTests):
    with subtests.test("no_auto_clip"):
        with pytest.raises(pydantic.ValidationError):
            KeypointAnnotation(keypoints=[(-2.1, 1.1, 0)])
        with pytest.raises(pydantic.ValidationError):
            KeypointAnnotation(keypoints=[(0.1, 2.1, 1)])
        with pytest.raises(pydantic.ValidationError):
            KeypointAnnotation(keypoints=[(0.1, 1.1, 2), (0.1, 2.1, 1)])

    with subtests.test("auto_clip"):
        kpt_ann = KeypointAnnotation(keypoints=[(-1.1, 1.1, 0)])
        assert 0 <= kpt_ann.keypoints[0][0] <= 1
        assert 0 <= kpt_ann.keypoints[0][1] <= 1
        kpt_ann = KeypointAnnotation(keypoints=[(0.1, 1.1, 1)])
        assert 0 <= kpt_ann.keypoints[0][0] <= 1
        assert 0 <= kpt_ann.keypoints[0][1] <= 1
        kpt_ann = KeypointAnnotation(keypoints=[(-2, 2, 2)])
        assert 0 <= kpt_ann.keypoints[0][0] <= 1
        assert 0 <= kpt_ann.keypoints[0][1] <= 1
    with subtests.test("numpy"):
        keypoints = KeypointAnnotation(keypoints=[(0.1, 0.2, 2)])
        assert np.allclose(keypoints.to_numpy(), np.array([0.1, 0.2, 2]))
        keypoints_list = [
            keypoints,
            KeypointAnnotation(keypoints=[(0.2, 0.3, 0)]),
            KeypointAnnotation(keypoints=[(0.3, 0.4, 1)]),
        ]
        assert np.allclose(
            KeypointAnnotation.combine_to_numpy(keypoints_list, ..., ...),
            np.array([[0.1, 0.2, 2], [0.2, 0.3, 0], [0.3, 0.4, 1]]),
        )


def test_segmentation_annotation(subtests: SubTests, tempdir: Path):
    mask = np.array(
        [
            [0, 1, 0, 0],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 1, 1],
        ]
    )
    with subtests.test("mask"):
        seg = SegmentationAnnotation(mask=mask)  # type: ignore
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"11213ON0"
        assert (
            seg.model_dump_json()
            == '{"height":4,"width":4,"counts":"11213ON0"}'
        )
        np.save(tempdir / "mask.npy", mask)
        seg = SegmentationAnnotation(mask=tempdir / "mask.npy")  # type: ignore
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"11213ON0"
        cv2.imwrite(str(tempdir / "mask.png"), mask)
        seg = SegmentationAnnotation(mask=tempdir / "mask.png")  # type: ignore
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"11213ON0"

    with subtests.test("polyline"):
        seg = SegmentationAnnotation(
            points=[(0, 0), (1, 0), (1, 1), (0, 1)],  # type: ignore
            height=4,
            width=4,
        )
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"0`0"
        seg_clipped = SegmentationAnnotation(
            points=[(-0.1, 0), (1.1, 0), (1, 1.5), (-0.6, 1)],  # type: ignore
            height=4,
            width=4,
        )
        assert seg == seg_clipped

        with pytest.raises(ValueError, match="must be integers"):
            SegmentationAnnotation(
                points=[(0, 0), (1, 0), (1, 1)],  # type: ignore
                height=4,
                width="4",
            )
        with pytest.raises(ValueError, match="2D points"):
            SegmentationAnnotation(
                points=[(0, 0, 0), (1, 0, 4)],  # type: ignore
                height=4,
                width=4,
            )

    with subtests.test("rle_bytes"):
        seg = SegmentationAnnotation(counts=b"11213ON0", height=4, width=4)
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"11213ON0"
        assert np.array_equal(seg.to_numpy(), mask)

    with subtests.test("rle_ints"):
        seg = SegmentationAnnotation(
            # counts are computed using FORTRAN order
            counts=[1, 1, 2, 2, 5, 1, 3, 1],  # type: ignore
            height=4,
            width=4,
        )
        assert seg.height == 4
        assert seg.width == 4
        assert seg.counts == b"11213ON0"
        assert np.array_equal(seg.to_numpy(), mask)

        with pytest.raises(ValueError, match="must be integers"):
            SegmentationAnnotation(
                counts=[1, 1, 2, 2, 5, 1, 3],  # type: ignore
                height=4,
                width="4",  # type: ignore
            )

    with subtests.test("numpy_simple"):
        masks = np.array(
            [
                [
                    [0, 1, 0, 0],
                    [1, 1, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 1, 1],
                ],
                [
                    [1, 0, 1, 1],
                    [0, 0, 1, 1],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                ],
            ]
        )

        annotations = [
            SegmentationAnnotation(mask=masks[i])  # type: ignore
            for i in range(len(masks))
        ]

        combined = SegmentationAnnotation.combine_to_numpy(
            annotations, [0, 1], 2
        )
        assert np.array_equal(combined, masks)

    with subtests.test("numpy_overlap"):
        masks = np.array(
            [
                [
                    [0, 1, 0, 0],
                    [1, 1, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 1, 1],
                ],
                [
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                ],
            ]
        )

        annotations = [
            SegmentationAnnotation(mask=masks[i])  # type: ignore
            for i in range(len(masks))
        ]

        combined = SegmentationAnnotation.combine_to_numpy(
            annotations, [0, 1], 2
        )
        assert np.array_equal(
            combined,
            np.array(
                [
                    [
                        [0, 1, 0, 0],
                        [1, 1, 0, 0],
                        [0, 0, 0, 0],
                        [0, 0, 1, 1],
                    ],
                    [
                        [1, 0, 1, 1],
                        [0, 0, 1, 1],
                        [0, 0, 0, 0],
                        [0, 0, 0, 0],
                    ],
                ]
            ),
        )

    with subtests.test("numpy_instance_segmentation"):
        masks = np.array(
            [
                [
                    [0, 1, 0, 0],
                    [1, 1, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 1, 1],
                ],
                [
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                ],
            ]
        )

        annotations = [
            InstanceSegmentationAnnotation(mask=masks[i])  # type: ignore
            for i in range(len(masks))
        ]

        combined = InstanceSegmentationAnnotation.combine_to_numpy(
            annotations, [0, 1], 2
        )
        assert np.array_equal(combined, masks)

    with subtests.test("invalid"):
        with pytest.raises(ValueError, match="Extra inputs"):
            SegmentationAnnotation(
                mask=mask,  # type: ignore
                points=[(1, 0), (2, 1)],  # type: ignore
            )

        with pytest.raises(ValueError, match="Unsupported mask format"):
            SegmentationAnnotation(mask="file.jpeg")  # type: ignore

        with pytest.raises(ValueError, match="Failed to load mask from image"):
            SegmentationAnnotation(mask="file.png")  # type: ignore

        with pytest.raises(TypeError, match="Mask must be either"):
            SegmentationAnnotation(mask=[1, 2, 3])  # type: ignore

        np.save(tempdir / "mask.npy", mask[None, None, ...])
        with pytest.raises(ValueError, match="Mask must be a 2D binary array"):
            SegmentationAnnotation(mask=tempdir / "mask.npy")  # type: ignore

        with pytest.raises(ValueError, match="at least 3 points"):
            SegmentationAnnotation(
                points=[(1, 0), (0, 1)],  # type: ignore
                height=4,
                width=4,
            )

        with pytest.raises(ValueError, match="outside of automatic clipping"):
            SegmentationAnnotation(
                points=[(-2.1, 0), (1.1, 0), (1, 1.5), (-0.6, 1)],  # type: ignore
                height=4,
                width=4,
            )

        with pytest.raises(ValueError, match="Field required"):
            SegmentationAnnotation(width=4)  # type: ignore

        with pytest.raises(
            ValueError, match="RLE counts must be a list of positive integers"
        ):
            SegmentationAnnotation(
                counts=[-1, 1, 2, 2],  # type: ignore
                height=4,
                width=4,
            )


def test_array_annotation(subtests: SubTests, tempdir: Path):
    arr = np.random.rand(100, 100)
    arr_path = tempdir / "array.npy"
    np.save(arr_path, arr)

    with subtests.test("simple"):
        annotation = ArrayAnnotation(path=arr_path)
        assert (
            ArrayAnnotation.model_validate_json(annotation.model_dump_json())
            == annotation
        )

    with subtests.test("numpy"):
        annotation = ArrayAnnotation(path=arr_path)
        assert np.array_equal(annotation.to_numpy(), arr)

        annotations = [ArrayAnnotation(path=arr_path) for _ in range(5)]
        array = ArrayAnnotation.combine_to_numpy(
            annotations, [0, 1, 2, 2, 1], 4
        )
        assert array.shape == (5, 4, 100, 100)
        assert np.allclose(array[0, 0, ...], arr)
        assert np.allclose(array[1, 1, ...], arr)
        assert np.allclose(array[2, 2, ...], arr)
        assert np.allclose(array[3, 2, ...], arr)
        assert np.allclose(array[4, 1, ...], arr)

    with subtests.test("invalid"):
        with pytest.raises(ValueError, match="Path does not"):
            ArrayAnnotation(path=Path("non_existent.npy"))

        cv2.imwrite(str(tempdir / "image.png"), np.zeros((100, 100, 3)))
        with pytest.raises(ValueError, match="must be a .npy file"):
            ArrayAnnotation(path=tempdir / "image.png")


def test_classification_annotation():
    arr = ClassificationAnnotation.combine_to_numpy(
        [ClassificationAnnotation() for _ in range(5)], [0, 1, 2, 2, 1], 6
    )
    assert np.array_equal(arr, np.array([1, 1, 1, 0, 0, 0]))


def test_detection(subtests: SubTests):
    with subtests.test("rescaling"):
        detection = Detection(
            **{
                "class": "person",
                "scale_to_boxes": True,
                "boundingbox": {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5},
                "keypoints": {
                    "keypoints": [(0.2, 0.4, 2), (0.5, 0.8, 2)],
                },
            }
        )

        assert detection.keypoints is not None
        assert detection.keypoints.keypoints == [
            (0.2 * 0.5 + 0.1, 0.4 * 0.5 + 0.2, 2),
            (0.5 * 0.5 + 0.1, 0.8 * 0.5 + 0.2, 2),
        ]

    with subtests.test("no_rescaling"):
        detection = Detection(
            **{
                "class": "person",
                "boundingbox": {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5},
                "keypoints": {
                    "keypoints": [(0.2, 0.4, 2), (0.5, 0.8, 2)],
                },
            }
        )

        assert detection.keypoints is not None
        assert detection.keypoints.keypoints == [(0.2, 0.4, 2), (0.5, 0.8, 2)]

    with subtests.test("invalid"), pytest.raises(
        ValueError, match="no bounding box is provided"
    ):
        Detection(
            **{
                "class": "person",
                "scale_to_boxes": True,
                "keypoints": {
                    "keypoints": [(0.2, 0.4, 2), (0.5, 0.8, 2)],
                },
            }
        )


def test_record(tempdir: Path):
    detection = Detection(
        **{
            "class": "person",
            "boundingbox": {"x": 0.1, "y": 0.2, "w": 0.5, "h": 0.5},
            "keypoints": {
                "keypoints": [(0.2, 0.4, 2), (0.5, 0.8, 2)],
            },
            "segmentation": {
                "mask": np.array(
                    [
                        [0, 1, 0, 0],
                        [1, 1, 0, 0],
                        [0, 0, 0, 0],
                        [0, 0, 1, 1],
                    ]
                ),
            },
            "instance_segmentation": {
                "mask": np.array(
                    [
                        [1, 1, 0, 0],
                        [1, 1, 0, 0],
                        [0, 0, 0, 0],
                        [0, 0, 0, 0],
                    ]
                ),
            },
            "metadata": {"age": 25},
            "sub_detections": {
                "head": {
                    "boundingbox": {
                        "x": 0.2,
                        "y": 0.3,
                        "w": 0.1,
                        "h": 0.1,
                    },
                }
            },
        }
    )
    filename = str(tempdir / "image.jpg")
    cv2.imwrite(filename, np.zeros((256, 256, 3), dtype=np.uint8))
    record = DatasetRecord(
        file=filename,  # type: ignore
        annotation=detection,
        task_name="test",
    )
    common = {
        "file": filename,
        "source_name": "image",
        "instance_id": -1,
    }
    expected_rows = [
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "boundingbox",
            "annotation": '{"x":0.1,"y":0.2,"w":0.5,"h":0.5}',
        },
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "keypoints",
            "annotation": '{"keypoints":[[0.2,0.4,2],[0.5,0.8,2]]}',
        },
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "segmentation",
            "annotation": '{"height":4,"width":4,"counts":"11213ON0"}',
        },
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "instance_segmentation",
            "annotation": '{"height":4,"width":4,"counts":"02208"}',
        },
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "metadata/age",
            "annotation": "25",
        },
        {
            **common,
            "task_name": "test",
            "class_name": "person",
            "task_type": "classification",
            "annotation": "{}",
        },
        {
            **common,
            "task_name": "test/head",
            "class_name": None,
            "task_type": "boundingbox",
            "annotation": '{"x":0.2,"y":0.3,"w":0.1,"h":0.1}',
        },
    ]
    assert list(record.to_parquet_rows()) == expected_rows



from importlib.metadata import version

import rich
import rich.box
import typer
from rich.markup import escape
from rich.table import Table

app = typer.Typer(
    name="Luxonis ML CLI",
    add_completion=True,
    pretty_exceptions_show_locals=False,
)

try:
    from luxonis_ml.data.__main__ import app as data_app

    app.add_typer(data_app, name="data", help="Dataset utilities.")
except ImportError:
    pass

try:
    from luxonis_ml.utils.__main__ import app as utils_app

    app.add_typer(utils_app, name="fs", help="Filesystem utilities.")
except ImportError:
    pass

try:
    from luxonis_ml.nn_archive.__main__ import app as utils_app

    app.add_typer(utils_app, name="archive", help="NN Archive utilities.")
except ImportError:
    pass


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"LuxonisML: {version('luxonis_ml')}")
        raise typer.Exit


@app.callback()
def main(
    _: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version and exit.",
    ),
):
    return


@app.command()
def checkhealth():
    """Check the health of the Luxonis ML library."""
    table = Table(
        title="Health Check",
        box=rich.box.ROUNDED,
    )
    table.add_column("Module", header_style="magenta i")
    table.add_column("Status", header_style="magenta i")
    table.add_column("Error", header_style="magenta i", max_width=50)
    for submodule in ["data", "utils", "nn_archive"]:
        error_message = ""
        try:
            __import__(f"luxonis_ml.{submodule}")
            status = "✅"
            style = "green"
        except ImportError as e:
            status = "❌"
            style = "red"
            error_message = escape(str(e.args[0]))
            if len(e.args) > 1:
                error_message += f" [bold]{escape(e.args[1])}[/bold]"

        table.add_row(submodule, status, error_message, style=style)

    console = rich.console.Console()
    console.print(table)


if __name__ == "__main__":
    app()



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np
from loguru import logger

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class SOLOParser(BaseParser):
    """Parses directory with SOLO annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── metadata.json
        │   ├── sensor_definitions.json
        │   ├── annotation_definitions.json
        │   ├── metric_definitions.json
        │   └── sequence.<SequenceNUM>/
        │       ├── step<StepNUM>.camera.jpg
        │       ├── step<StepNUM>.frame_data.json
        │       └── (OPTIONAL: step<StepNUM>.camera.semantic segmentation.jpg)
        ├── valid/
        └── test/

    This is the default format returned by Unity simulation engine.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        """Validates if a split subdirectory is in an expected format.

        @type split_path: Path
        @param split_path: Path to split directory.
        @rtype: Optional[Dict[str, Any]]
        @return: Dictionary with kwargs to pass to L{from_split} method
            or C{None} if the split is not in the expected format.
        """
        if not split_path.exists():
            return None
        # check if all json files are present
        for json_fname in [
            "annotation_definitions.json",
            "metadata.json",
            "metric_definitions.json",
            "sensor_definitions.json",
        ]:
            json_path = next(split_path.glob(json_fname), None)
            if not json_path:
                return None
        with open(split_path / "metadata.json") as json_file:
            metadata_dict = json.load(json_file)
        # check if all sequences are present
        total_sequences_expected = metadata_dict["totalSequences"]
        total_sequences = len(
            [d for d in split_path.glob("sequence*") if d.is_dir()]
        )
        if total_sequences != total_sequences_expected:
            logger.warning(
                f"Expected {total_sequences_expected} based on metadata.json, "
                f"but found {total_sequences} sequences."
            )
        return {"split_path": split_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        """Validates if the dataset is in an expected format.

        @type dataset_dir: Path
        @param dataset_dir: Path to source dataset directory.
        @rtype: bool
        @return: True if the dataset is in the expected format.
        """
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if SOLOParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        """Parses all present data to L{LuxonisDataset} format.

        @type dataset_dir: str
        @param dataset_dir: Path to source dataset directory.
        @rtype: Tuple[List[Path], List[Path], List[Path]]
        @return: Tuple with added images for train, valid and test
            splits.
        """

        added_train_imgs = self._parse_split(split_path=dataset_dir / "train")
        added_valid_imgs = self._parse_split(split_path=dataset_dir / "valid")
        added_test_imgs = self._parse_split(split_path=dataset_dir / "test")

        return added_train_imgs, added_valid_imgs, added_test_imgs

    def from_split(self, split_path: Path) -> ParserOutput:
        """Parses data in a split subdirectory from SOLO format to
        L{LuxonisDataset} format.

        @type split_path: Path
        @param split_path: Path to directory with sequences of images
            and annotations.
        @rtype: L{ParserOutput}
        @return: C{LuxonisDataset} generator, list of class names,
            skeleton dictionary for keypoints and list of added images.
        """

        if not split_path.exists():
            raise FileNotFoundError(f"{split_path} path non-existent.")

        annotation_definitions_path = (
            split_path / "annotation_definitions.json"
        )
        if annotation_definitions_path.exists():
            with open(annotation_definitions_path) as json_file:
                annotation_definitions_dict = json.load(json_file)
        else:
            raise FileNotFoundError(
                f"{annotation_definitions_path} path non-existent."
            )

        bbox_class_names = self._get_solo_bbox_class_names(
            annotation_definitions_dict
        )

        if not bbox_class_names:
            raise ValueError("No class_names identified. ")

        keypoint_labels = self._get_solo_keypoint_names(
            annotation_definitions_dict
        )

        skeletons = {
            class_name: {"labels": keypoint_labels}
            for class_name in bbox_class_names
        }

        def generator() -> DatasetIterator:
            for sequence_path in split_path.glob("sequence*"):
                processed_annotations_per_step: Dict[
                    str, set
                ] = {}  # Seperate json files can have the same annotations in them
                for frame_path in sequence_path.glob("*.frame_data*.json"):
                    frame = json.loads(frame_path.read_text())

                    curent_step = frame["step"]
                    if curent_step not in processed_annotations_per_step:
                        processed_annotations_per_step[curent_step] = set()

                    for capture in frame.get("captures", []):
                        img_fname = capture["filename"]
                        img_w, img_h = capture["dimension"]
                        annotations = capture["annotations"]
                        img_path = sequence_path / img_fname
                        if not img_path.exists():
                            raise FileNotFoundError(
                                f"{img_path} not existent."
                            )
                        instance_segmentations = {}
                        instance_keypoints = {}
                        bounding_boxes = {}
                        for anno in annotations:
                            if (
                                "SemanticSegmentationAnnotation"
                                not in processed_annotations_per_step[
                                    curent_step
                                ]
                                and anno["@type"].endswith(
                                    "SemanticSegmentationAnnotation"
                                )
                            ):
                                processed_annotations_per_step[
                                    curent_step
                                ].add("SemanticSegmentationAnnotation")

                                mask_fname = anno["filename"]
                                mask_path = sequence_path / mask_fname
                                mask = cv2.imread(mask_path)

                                mask_int = (
                                    (mask[..., 0].astype(np.uint32) << 16)
                                    | (mask[..., 1].astype(np.uint32) << 8)
                                    | mask[..., 2].astype(np.uint32)
                                )

                                for instance in anno.get("instances", []):
                                    class_name = instance["labelName"]
                                    r, g, b, _ = instance["pixelValue"]
                                    target_int = (b << 16) | (g << 8) | r
                                    curr_mask = (
                                        mask_int == target_int
                                    ).astype(np.uint8)
                                    yield {
                                        "file": img_path,
                                        "annotation": {
                                            "class": class_name,
                                            "segmentation": {
                                                "mask": curr_mask,
                                            },
                                        },
                                    }

                            elif (
                                "BoundingBox2DAnnotation"
                                not in processed_annotations_per_step[
                                    curent_step
                                ]
                                and anno["@type"].endswith(
                                    "BoundingBox2DAnnotation"
                                )
                            ):
                                processed_annotations_per_step[
                                    curent_step
                                ].add("BoundingBox2DAnnotation")
                                bbox_annotations = anno.get("values", [])

                                for bbox_annotation in bbox_annotations:
                                    class_name = bbox_annotation["labelName"]
                                    origin = bbox_annotation["origin"]
                                    dimension = bbox_annotation["dimension"]
                                    xmin, ymin = origin
                                    bbox_w, bbox_h = dimension

                                    instance_id = bbox_annotation["instanceId"]
                                    bounding_boxes[instance_id] = {
                                        "file": img_path,
                                        "annotation": {
                                            "class": class_name,
                                            "instance_id": instance_id,
                                            "boundingbox": {
                                                "x": xmin / img_w,
                                                "y": ymin / img_h,
                                                "w": bbox_w / img_w,
                                                "h": bbox_h / img_h,
                                            },
                                        },
                                    }

                            elif (
                                "InstanceSegmentationAnnotation"
                                not in processed_annotations_per_step[
                                    curent_step
                                ]
                                and anno["@type"].endswith(
                                    "InstanceSegmentationAnnotation"
                                )
                            ):
                                processed_annotations_per_step[
                                    curent_step
                                ].add("InstanceSegmentationAnnotation")

                                mask_fname = anno["filename"]
                                mask_path = sequence_path / mask_fname
                                mask = cv2.imread(mask_path)

                                mask_int = (
                                    (mask[..., 0].astype(np.uint32) << 16)
                                    | (mask[..., 1].astype(np.uint32) << 8)
                                    | mask[..., 2].astype(np.uint32)
                                )

                                for instance in anno.get("instances", []):
                                    r, g, b, _ = instance["color"]
                                    target_int = (b << 16) | (g << 8) | r
                                    curr_mask = (
                                        mask_int == target_int
                                    ).astype(np.uint8)
                                    instance_id = instance["instanceId"]

                                    instance_segmentations[instance_id] = {
                                        "file": img_path,
                                        "annotation": {
                                            "instance_id": instance_id,
                                            "instance_segmentation": {
                                                "mask": curr_mask,
                                            },
                                        },
                                    }

                            elif (
                                "KeypointAnnotation"
                                not in processed_annotations_per_step[
                                    curent_step
                                ]
                                and anno["@type"].endswith(
                                    "KeypointAnnotation"
                                )
                            ):
                                processed_annotations_per_step[
                                    curent_step
                                ].add("KeypointAnnotation")
                                keypoint_annotations = anno.get("values", [])

                                for (
                                    keypoints_annotation
                                ) in keypoint_annotations:
                                    keypoints = []
                                    for keypoint in keypoints_annotation[
                                        "keypoints"
                                    ]:
                                        x, y = keypoint["location"]
                                        visibility = keypoint["state"]
                                        keypoints.append(
                                            (x / img_w, y / img_h, visibility)
                                        )

                                    instance_id = keypoints_annotation[
                                        "instanceId"
                                    ]

                                    instance_keypoints[instance_id] = {
                                        "file": img_path,
                                        "annotation": {
                                            "instance_id": instance_id,
                                            "keypoints": {
                                                "keypoints": keypoints,
                                            },
                                        },
                                    }
                        # Hard dependencies between bbox, keypoints and instance_segmentations
                        non_empty_annotations = []
                        if bounding_boxes:
                            non_empty_annotations.append(bounding_boxes)
                        if instance_keypoints:
                            non_empty_annotations.append(instance_keypoints)
                        if instance_segmentations:
                            non_empty_annotations.append(
                                instance_segmentations
                            )

                        if non_empty_annotations:
                            common_instance_ids = set.intersection(
                                *[
                                    set(ann.keys())
                                    for ann in non_empty_annotations
                                ]
                            )
                        else:
                            common_instance_ids = set()

                        for instance_id in common_instance_ids:
                            annotation_entry = {
                                "class": bounding_boxes[instance_id][
                                    "annotation"
                                ]["class"],
                                "instance_id": instance_id,
                                "boundingbox": bounding_boxes[instance_id][
                                    "annotation"
                                ]["boundingbox"],
                            }
                            if instance_keypoints:
                                annotation_entry["keypoints"] = (
                                    instance_keypoints[instance_id][
                                        "annotation"
                                    ]["keypoints"]
                                )
                            if instance_segmentations:
                                annotation_entry["instance_segmentation"] = (
                                    instance_segmentations[instance_id][
                                        "annotation"
                                    ]["instance_segmentation"]
                                )

                            yield {
                                "file": img_path,
                                "annotation": annotation_entry,
                            }

        return generator(), skeletons, []

    def _get_solo_annotation_types(
        self, annotation_definitions_dict: Dict[str, Any]
    ) -> List[str]:
        """List all annotation types present in the dataset.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of annotation types (e.g. ["BoundingBox2DAnnotation", "SemanticSegmentationAnnotation", ...]).
        """

        annotation_types = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_types.append(
                definition["@type"].replace("type.unity.com/unity.solo.", "")
            )
        return annotation_types

    def _get_solo_bbox_class_names(
        self, annotation_definitions_dict: Dict[str, Any]
    ) -> List[str]:
        """List class names for BoundingBox2DAnnotation type.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of class names (e.g. ["car", "motorbike", ...]).
        """

        class_names = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_type = definition["@type"].replace(
                "type.unity.com/unity.solo.", ""
            )
            if annotation_type == "BoundingBox2DAnnotation":
                names = [spec["label_name"] for spec in definition["spec"]]
                ids = [spec["label_id"] for spec in definition["spec"]]
                class_names = [c for _, c in sorted(zip(ids, names))]
        return class_names

    def _get_solo_keypoint_names(
        self, annotation_definitions_dict: Dict[str, Any]
    ) -> List[str]:
        """List keypoint labels for all classes.

        @type annotation_definitions_dict: dict
        @param annotation_definitions_dict: annotation_definitions.json read as dict.
        @rtype: List[str]
        @return: List of keypoint labels (e.g. ["wheel_front_motorbike", "wheel_back_motorbike", ...]).
        """

        keypoint_labels = []
        for definition in annotation_definitions_dict["annotationDefinitions"]:
            annotation_type = definition["@type"].replace(
                "type.unity.com/unity.solo.", ""
            )
            if annotation_type == "KeypointAnnotation":
                keypoints = definition["template"]["keypoints"]
                labels = [keypoint["label"] for keypoint in keypoints]
                ids = [keypoint["index"] for keypoint in keypoints]
                keypoint_labels = [c for _, c in sorted(zip(ids, labels))]
        return keypoint_labels



from contextlib import contextmanager
from typing import Generator


@contextmanager
def guard_missing_extra(name: str) -> Generator[None, None, None]:
    try:
        yield
    except ImportError as e:
        raise ImportError(
            f"Error importing `luxonis-ml.{name}`. This can mean that some of the dependencies of `luxonis-ml[{name}]` are not installed. "
            f"Ensure you installed the package with the `[{name}]` or `[all]` extra specified. "
            f"Use `pip install luxonis-ml[{name}]` to install dependencies for the `{name}` submodule.",
            str(e),
        ) from e


__all__ = ["guard_missing_extra"]



import json
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
)

import cv2
import numpy as np
import pycocotools.mask
from loguru import logger
from PIL import Image, ImageDraw
from pydantic import (
    Field,
    GetCoreSchemaHandler,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.types import FilePath, PositiveInt
from pydantic_core import core_schema
from typing_extensions import Annotated, Self, TypeAlias, override

from luxonis_ml.data.utils.parquet import ParquetRecord
from luxonis_ml.typing import check_type
from luxonis_ml.utils import BaseModelExtraForbid

KeypointVisibility: TypeAlias = Literal[0, 1, 2]
NormalizedFloat: TypeAlias = Annotated[float, Field(ge=0, le=1)]


class Category(str):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)


class Detection(BaseModelExtraForbid):
    class_name: Optional[str] = Field(None, alias="class")
    instance_id: int = -1

    metadata: Dict[str, Union[int, float, str, Category]] = {}

    boundingbox: Optional["BBoxAnnotation"] = None
    keypoints: Optional["KeypointAnnotation"] = None
    instance_segmentation: Optional["InstanceSegmentationAnnotation"] = None
    segmentation: Optional["SegmentationAnnotation"] = None
    array: Optional["ArrayAnnotation"] = None

    scale_to_boxes: bool = False

    sub_detections: Dict[str, "Detection"] = {}

    @model_validator(mode="after")
    def validate_names(self) -> Self:
        for name in self.sub_detections:
            check_valid_identifier(name, label="Sub-detection name")
        for key in self.metadata:
            check_valid_identifier(key, label="Metadata key")
        return self

    @model_validator(mode="after")
    def rescale_values(self) -> Self:
        if not self.scale_to_boxes:
            return self
        if self.boundingbox is None:
            raise ValueError(
                "`scaled_to_boxes` is set to True, "
                "but no bounding box is provided."
            )
        x, y, w, h = (
            self.boundingbox.x,
            self.boundingbox.y,
            self.boundingbox.w,
            self.boundingbox.h,
        )

        if self.keypoints is not None:
            self.keypoints = KeypointAnnotation(
                keypoints=[
                    (x + w * kp[0], y + h * kp[1], kp[2])
                    for kp in self.keypoints.keypoints
                ]
            )
        return self

    def get_task_types(self) -> Set[str]:
        task_types = {
            task_type
            for task_type in [
                "boundingbox",
                "keypoints",
                "segmentation",
                "instance_segmentation",
                "array",
            ]
            if getattr(self, task_type) is not None
        }
        if self.class_name is not None:
            task_types.add("classification")
        for metadata_key in self.metadata:
            task_types.add(f"metadata/{metadata_key}")

        return task_types


class Annotation(ABC, BaseModelExtraForbid):
    """Base class for an annotation."""

    @staticmethod
    @abstractmethod
    def combine_to_numpy(
        annotations: List["Annotation"], classes: List[int], n_classes: int
    ) -> np.ndarray: ...


class ClassificationAnnotation(Annotation):
    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["ClassificationAnnotation"],
        classes: List[int],
        n_classes: int,
    ) -> np.ndarray:
        classify_vector = np.zeros(n_classes)
        for i in range(len(annotations)):
            classify_vector[classes[i]] = 1
        return classify_vector


class BBoxAnnotation(Annotation):
    """Bounding box annotation.

    Values are normalized based on the image size.

    @type x: float
    @ivar x: The top-left x coordinate of the bounding box. Normalized
        to M{[0, 1]}.
    @type y: float
    @ivar y: The top-left y coordinate of the bounding box. Normalized
        to M{[0, 1]}.
    @type w: float
    @ivar w: The width of the bounding box. Normalized to M{[0, 1]}.
    @type h: float
    @ivar h: The height of the bounding box. Normalized to M{[0, 1]}.
    """

    x: NormalizedFloat
    y: NormalizedFloat
    w: NormalizedFloat
    h: NormalizedFloat

    def to_numpy(self, class_id: int) -> np.ndarray:
        return np.array([class_id, self.x, self.y, self.w, self.h])

    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["BBoxAnnotation"],
        classes: List[int],
        n_classes: int = ...,
    ) -> np.ndarray:
        boxes = np.empty((len(annotations), 5))
        for i, ann in enumerate(annotations):
            boxes[i] = ann.to_numpy(classes[i])
        return boxes

    @model_validator(mode="before")
    @classmethod
    def validate_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        warn = False
        for key in ["x", "y", "w", "h"]:
            if values[key] < -2 or values[key] > 2:
                raise ValueError(
                    "BBox annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            if not (0 <= values[key] <= 1):
                warn = True
                values[key] = max(0, min(1, values[key]))
        if warn:
            logger.warning(
                "BBox annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )

        return cls._clip_sum(values)

    @staticmethod
    def _clip_sum(values: Dict[str, Any]) -> Dict[str, Any]:
        if values["x"] + values["w"] > 1:
            values["w"] = 1 - values["x"]
            logger.warning(
                "BBox annotation has x + width > 1. Clipping width so the sum is 1."
            )
        if values["y"] + values["h"] > 1:
            values["h"] = 1 - values["y"]
            logger.warning(
                "BBox annotation has y + height > 1. Clipping height so the sum is 1."
            )
        return values


class KeypointAnnotation(Annotation):
    """Keypoint annotation.

    Values are normalized to M{[0, 1]} based on the image size.

    @type keypoints: List[Tuple[float, float, L{KeypointVisibility}]]
    @ivar keypoints: List of keypoints. Each keypoint is a tuple of (x, y, visibility).
        x and y are normalized to M{[0, 1]}. visibility is one of M{0}, M{1}, or M{2} where:
            - 0: Not visible / not labeled
            - 1: Occluded
            - 2: Visible
    """

    keypoints: List[
        Tuple[NormalizedFloat, NormalizedFloat, KeypointVisibility]
    ]

    def to_numpy(self) -> np.ndarray:
        return np.array(self.keypoints).reshape((-1, 3)).flatten()

    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["KeypointAnnotation"],
        classes: List[int] = ...,
        n_classes: int = ...,
    ) -> np.ndarray:
        keypoints = np.empty(
            (len(annotations), len(annotations[0].keypoints) * 3)
        )
        for i, ann in enumerate(annotations):
            keypoints[i] = ann.to_numpy()
        return keypoints

    @model_validator(mode="before")
    @classmethod
    def validate_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "keypoints" not in values:
            return values

        warn = False
        for i, keypoint in enumerate(values["keypoints"]):
            if (keypoint[0] < -2 or keypoint[0] > 2) or (
                keypoint[1] < -2 or keypoint[1] > 2
            ):
                raise ValueError(
                    "Keypoint annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            new_keypoint = list(keypoint)
            if not (0 <= keypoint[0] <= 1):
                new_keypoint[0] = max(0, min(1, keypoint[0]))
                warn = True
            if not (0 <= keypoint[1] <= 1):
                new_keypoint[1] = max(0, min(1, keypoint[1]))
                warn = True
            values["keypoints"][i] = tuple(new_keypoint)

        if warn:
            logger.warning(
                "Keypoint annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )
        return values


class SegmentationAnnotation(Annotation):
    """Run-length encoded segmentation mask.

    @type height: int
    @ivar height: The height of the segmentation mask.

    @type width: int
    @ivar width: The width of the segmentation mask.

    @type counts: Union[List[int], bytes]
    @ivar counts: The run-length encoded mask.
        This can be a list of integers or a byte string.

    @see: U{Run-length encoding<https://en.wikipedia.org/wiki/Run-length_encoding>}
    """

    height: PositiveInt
    width: PositiveInt
    counts: bytes

    def to_numpy(self) -> np.ndarray:
        with warnings.catch_warnings(record=True):
            return pycocotools.mask.decode(
                {"counts": self.counts, "size": [self.height, self.width]}
            ).astype(np.uint8)

    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["SegmentationAnnotation"],
        classes: List[int],
        n_classes: int,
    ) -> np.ndarray:
        ref = annotations[0]
        width, height = ref.width, ref.height
        masks = np.stack([ann.to_numpy() for ann in annotations])

        segmentation = np.zeros((n_classes, height, width), dtype=np.uint8)

        assigned_pixels = np.zeros((height, width), dtype=bool)
        for i, class_id in enumerate(classes):
            mask = masks[i] & (assigned_pixels == 0)
            segmentation[class_id, ...] = np.maximum(
                segmentation[class_id, ...], mask
            )
            assigned_pixels |= mask.astype(bool)

        return segmentation

    @field_serializer("counts", when_used="json")
    def serialize_counts(self, counts: bytes) -> str:
        return counts.decode("utf-8")

    @model_validator(mode="before")
    @classmethod
    def validate_rle(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if {"counts", "width", "height"} - set(values.keys()):
            return values

        height = values["height"]
        width = values["width"]

        if not check_type(height, int) or not check_type(width, int):
            raise ValueError("Height and width must be integers")

        counts = values["counts"]
        if isinstance(counts, str):
            values["counts"] = counts.encode("utf-8")

        elif isinstance(counts, list):
            for c in counts:
                if not isinstance(c, int) or c < 0:
                    raise ValueError(
                        "RLE counts must be a list of positive integers"
                    )

            with warnings.catch_warnings(record=True):
                rle: Any = pycocotools.mask.frPyObjects(
                    {"counts": counts, "size": [height, width]},  # type: ignore
                    height,
                    width,
                )
            values["counts"] = rle["counts"]
            values["height"] = rle["size"][0]
            values["width"] = rle["size"][1]

        return values

    @staticmethod
    def _numpy_to_rle(mask: np.ndarray) -> Dict[str, Any]:
        mask = np.asfortranarray(mask.astype(np.uint8))
        with warnings.catch_warnings(record=True):
            rle = pycocotools.mask.encode(mask)
        return {
            "height": rle["size"][0],
            "width": rle["size"][1],
            "counts": rle["counts"].decode("utf-8"),  # type: ignore
        }

    @model_validator(mode="before")
    @classmethod
    def validate_mask(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "mask" not in values:
            return values
        values = deepcopy(values)

        mask = values.pop("mask")
        if isinstance(mask, (str, Path)):
            mask_path = Path(mask)
            if mask_path.suffix == ".npy":
                try:
                    mask = np.load(mask_path)
                except Exception as e:
                    raise ValueError(
                        f"Failed to load mask from array at '{mask_path}'"
                    ) from e
            elif mask_path.suffix == ".png":
                try:
                    mask = (
                        cv2.imread(str(mask_path), cv2.IMREAD_GRAYSCALE)
                        .astype(bool)
                        .astype(np.uint8)
                    )
                except Exception as e:
                    raise ValueError(
                        f"Failed to load mask from image at '{mask_path}'"
                    ) from e
            else:
                raise ValueError(
                    f"Unsupported mask format: {mask_path.suffix}. "
                    "Supported formats are .npy and .png"
                )
        if not isinstance(mask, np.ndarray):
            raise TypeError(
                "Mask must be either a numpy array, "
                "or a path to a saved numpy array"
            )

        if mask.ndim != 2:
            raise ValueError("Mask must be a 2D binary array")

        return {**cls._numpy_to_rle(mask), **values}

    @model_validator(mode="before")
    @classmethod
    def validate_polyline(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if {"points", "width", "height"} - set(values.keys()):
            return values

        values = deepcopy(values)

        width = values.pop("width")
        height = values.pop("height")
        if not check_type(height, int) or not check_type(width, int):
            raise ValueError("Height and width must be integers")

        points = values.pop("points")
        if not check_type(points, List[Tuple[float, float]]):
            raise ValueError("Polyline must be a list of float 2D points")

        if len(points) < 3:
            raise ValueError("Polyline must contain at least 3 points")

        cls._clip_points(points)

        polyline = [(round(x * width), round(y * height)) for x, y in points]
        mask = Image.new("L", (width, height), 0)
        draw = ImageDraw.Draw(mask)
        draw.polygon(polyline, fill=1, outline=1)
        return {"mask": np.array(mask).astype(np.uint8), **values}

    @staticmethod
    def _clip_points(points: List[Tuple[float, float]]) -> None:
        warn = False
        for i in range(len(points)):
            x, y = points[i]
            if (x < -2 or x > 2) or (y < -2 or y > 2):
                raise ValueError(
                    "Polyline annotation has value outside of automatic clipping range ([-2, 2]). "
                    "Values should be normalized based on image size to range [0, 1]."
                )
            new_x, new_y = x, y
            if not (0 <= x <= 1):
                new_x = max(0, min(1, x))
                warn = True
            if not (0 <= y <= 1):
                new_y = max(0, min(1, y))
                warn = True

            points[i] = (new_x, new_y)

        if warn:
            logger.warning(
                "Polyline annotation has values outside of [0, 1] range. Clipping them to [0, 1]."
            )


class InstanceSegmentationAnnotation(SegmentationAnnotation):
    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["InstanceSegmentationAnnotation"],
        classes: List[int] = ...,
        n_classes: int = ...,
    ) -> np.ndarray:
        return np.stack([ann.to_numpy() for ann in annotations])


class ArrayAnnotation(Annotation):
    """A custom unspecified annotation that is an arbitrary numpy array.

    All instances of this annotation must have the same shape.

    @type path: FilePath
    @ivar path: The path to the numpy array saved as a C{.npy} file.
    """

    path: FilePath

    def to_numpy(self) -> np.ndarray:
        return np.load(self.path)

    @staticmethod
    @override
    def combine_to_numpy(
        annotations: List["ArrayAnnotation"],
        classes: List[int],
        n_classes: int,
    ) -> np.ndarray:
        out_arr = np.zeros(
            (len(annotations), n_classes, *np.load(annotations[0].path).shape)
        )
        for i, ann in enumerate(annotations):
            out_arr[i, classes[i]] = ann.to_numpy()
        return out_arr

    @field_serializer("path", when_used="json")
    def serialize_path(self, value: FilePath) -> str:
        return str(value)

    @field_validator("path")
    @classmethod
    def validate_path(cls, path: FilePath) -> FilePath:
        if path.suffix != ".npy":
            raise ValueError(
                f"Array annotation file must be a .npy file. Got {path}"
            )
        try:
            np.load(path)
        except Exception as e:
            raise ValueError(
                f"Failed to load array annotation from {path}."
            ) from e
        return path


class DatasetRecord(BaseModelExtraForbid):
    files: Dict[str, FilePath]
    annotation: Optional[Detection] = None
    task_name: str = ""

    @property
    def file(self) -> FilePath:
        if len(self.files) != 1:
            raise ValueError("DatasetRecord must have exactly one file")
        return next(iter(self.files.values()))

    @model_validator(mode="after")
    def validate_task_name_valid_identifier(self) -> Self:
        check_valid_identifier(self.task_name, label="Task name")
        return self

    @model_validator(mode="before")
    @classmethod
    def validate_task_name(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if "task" in values:
            warnings.warn(
                "The 'task' field is deprecated. Use 'task_name' instead.",
                stacklevel=2,
            )
            values["task_name"] = values.pop("task")
        return values

    @model_validator(mode="before")
    @classmethod
    def validate_files(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        values = deepcopy(values)
        if "file" in values:
            values["files"] = {"image": values.pop("file")}
        return values

    def to_parquet_rows(self) -> Iterable[ParquetRecord]:
        """Converts an annotation to a dictionary for writing to a
        parquet file.

        @rtype: L{ParquetDict}
        @return: A dictionary of annotation data.
        """
        yield from self._to_parquet_rows(self.annotation, self.task_name)

    def _to_parquet_rows(
        self, annotation: Optional[Detection], task_name: str
    ) -> Iterable[ParquetRecord]:
        """Converts an annotation to a dictionary for writing to a
        parquet file.

        @rtype: L{ParquetDict}
        @return: A dictionary of annotation data.
        """
        for source, file_path in self.files.items():
            if annotation is None:
                yield {
                    "file": str(file_path),
                    "source_name": source,
                    "task_name": task_name,
                    "class_name": None,
                    "instance_id": None,
                    "task_type": None,
                    "annotation": None,
                }
            else:
                for task_type in [
                    "boundingbox",
                    "keypoints",
                    "segmentation",
                    "instance_segmentation",
                    "array",
                ]:
                    label: Optional[Annotation] = getattr(
                        annotation, task_type
                    )

                    if label is not None:
                        yield {
                            "file": str(file_path),
                            "source_name": source,
                            "task_name": task_name,
                            "class_name": annotation.class_name,
                            "instance_id": annotation.instance_id,
                            "task_type": task_type,
                            "annotation": label.model_dump_json(),
                        }
                for key, data in annotation.metadata.items():
                    yield {
                        "file": str(file_path),
                        "source_name": source,
                        "task_name": task_name,
                        "class_name": annotation.class_name,
                        "instance_id": annotation.instance_id,
                        "task_type": f"metadata/{key}",
                        "annotation": json.dumps(data),
                    }
                if annotation.class_name is not None:
                    yield {
                        "file": str(file_path),
                        "source_name": source,
                        "task_name": task_name,
                        "class_name": annotation.class_name,
                        "instance_id": annotation.instance_id,
                        "task_type": "classification",
                        "annotation": "{}",
                    }
                for name, detection in annotation.sub_detections.items():
                    yield from self._to_parquet_rows(
                        detection, f"{task_name}/{name}"
                    )


def check_valid_identifier(name: str, *, label: str) -> None:
    """Check if a name is a valid Python identifier after converting
    dashes to underscores.

    Albumentations requires that the names of the targets
    passed as `additional_targets` are valid Python identifiers.
    """
    name = name.replace("-", "_")
    if name and not name.isidentifier():
        raise ValueError(
            f"{label} can only contain alphanumeric characters, "
            "underscores, and dashes. Additionaly, the first character "
            f"must be a letter or underscore. Got {name}"
        )


def load_annotation(task_type: str, data: Dict[str, Any]) -> "Annotation":
    classes = {
        "classification": ClassificationAnnotation,
        "boundingbox": BBoxAnnotation,
        "keypoints": KeypointAnnotation,
        "segmentation": SegmentationAnnotation,
        "instance_segmentation": InstanceSegmentationAnnotation,
        "array": ArrayAnnotation,
    }
    if task_type not in classes:
        raise ValueError(f"Unknown label type: {task_type}")
    return classes[task_type](**data)



import json
import uuid
from pathlib import Path
from typing import List

import cv2
import numpy as np

from luxonis_ml.data import BucketStorage, LuxonisDataset, LuxonisLoader
from luxonis_ml.typing import Params
from luxonis_ml.utils import LuxonisFileSystem

from .utils import gather_tasks


def get_annotations(sequence_path: Path):
    frame_data = sequence_path / "step0.frame_data.json"
    with open(frame_data) as f:
        data = json.load(f)["captures"][0]
        frame_data = data["annotations"]

    return {anno["@type"].split(".")[-1]: anno for anno in frame_data}


def test_parking_lot_generate(
    tempdir: Path,
    bucket_storage: BucketStorage,
    dataset_name: str,
    augmentation_config: List[Params],
    height: int,
    width: int,
    storage_url: str,
):
    data_path = LuxonisFileSystem.download(
        f"{storage_url}/D2_ParkingLot", tempdir
    )
    dataset = LuxonisDataset(
        dataset_name,
        delete_local=True,
        bucket_storage=bucket_storage,
        delete_remote=True,
    )
    dataset.add(generator(data_path, tempdir))
    dataset.make_splits((0.8, 0.1, 0.1))
    assert gather_tasks(dataset) == {
        "car/array",
        "car/boundingbox",
        "car/classification",
        "car/instance_segmentation",
        "car/keypoints",
        "car/metadata/brand",
        "car/metadata/color",
        "color/classification",
        "color/segmentation",
        "motorbike/array",
        "motorbike/boundingbox",
        "motorbike/classification",
        "motorbike/instance_segmentation",
        "motorbike/keypoints",
        "motorbike/metadata/brand",
        "motorbike/metadata/color",
    }
    accumulated_tasks = set()
    loader = LuxonisLoader(
        dataset,
        augmentation_config=augmentation_config,
        height=height,
        width=width,
        color_space="BGR",
        keep_aspect_ratio=True,
    )
    for _, labels in loader:
        accumulated_tasks.update(labels.keys())

    assert accumulated_tasks == gather_tasks(dataset)


# TODO: Simplify the dataset so the code can be cleaner
def generator(base_path: Path, tempdir: Path):
    array_counter = 0
    seen = set()
    batch = []
    for sequence_path in base_path.glob("sequence.*"):
        filepath = sequence_path / "step0.camera.jpg"
        if not filepath.exists():
            filepath = sequence_path / "step0.camera_0.jpg"

        file_hash_uuid = str(
            uuid.uuid5(uuid.NAMESPACE_URL, filepath.read_bytes().hex())
        )
        if file_hash_uuid in seen:
            continue
        seen.add(file_hash_uuid)
        annotations = get_annotations(sequence_path)
        W, H = annotations["SemanticSegmentationAnnotation"]["dimension"]
        bbox = annotations["BoundingBox2DAnnotation"]["values"][0]
        instance_id = bbox["instanceId"]
        x, y = bbox["origin"]
        w, h = bbox["dimension"]
        label_name: str = bbox["labelName"]
        *brand, color, vehicle_type = label_name.split("-")
        brand = "-".join(brand)
        vehicle_type = vehicle_type.lower()
        if vehicle_type == "motorbiek" or color not in [
            "RED",
            "BLUE",
            "GREEN",
        ]:
            continue

        keypoints = []
        kp_annotations = annotations["KeypointAnnotation"]["values"][0][
            "keypoints"
        ]
        if vehicle_type == "motorbike":
            kp_annotations = kp_annotations[:3]
        else:
            kp_annotations = kp_annotations[3:]

        for kp in kp_annotations:
            kpt_x, kpt_y = kp["location"]
            state = kp["state"]
            if vehicle_type == "motorbike":
                state = 2
            keypoints.append([kpt_x / W, kpt_y / H, state])

        mask_path = annotations["SemanticSegmentationAnnotation"]["filename"]
        mask = cv2.imread(
            str(sequence_path / mask_path), cv2.IMREAD_GRAYSCALE
        ).astype(bool)

        batch.append(
            {
                "filepath": filepath,
                "W": W,
                "H": H,
                "x": x,
                "y": y,
                "w": w,
                "h": h,
                "vehicle_type": vehicle_type,
                "brand": brand,
                "color": color,
                "instance_id": instance_id,
                "mask": mask,
                "keypoints": keypoints,
            }
        )

        if len(batch) == 4:
            images = [cv2.imread(str(item["filepath"])) for item in batch]
            H_img, W_img, C = images[0].shape
            combined_image = np.empty(
                (2 * H_img, 2 * W_img, C), dtype=images[0].dtype
            )
            color_mask = np.zeros((3, 2 * H_img, 2 * W_img), dtype=np.uint8)

            positions = [(0, 0), (1, 0), (0, 1), (1, 1)]
            annotations_list = []

            for idx, item in enumerate(batch):
                i, j = positions[idx]
                combined_image[
                    j * H_img : (j + 1) * H_img, i * W_img : (i + 1) * W_img, :
                ] = images[idx]

                x_orig = item["x"] / item["W"]
                y_orig = item["y"] / item["H"]
                w_orig = item["w"] / item["W"]
                h_orig = item["h"] / item["H"]

                x_new = (x_orig + i) / 2
                y_new = (y_orig + j) / 2
                w_new = w_orig / 2
                h_new = h_orig / 2

                for kp in item["keypoints"]:
                    if kp[2] != 0:
                        kp[0] = (kp[0] + i) / 2
                        kp[1] = (kp[1] + j) / 2

                H_mask, W_mask = item["mask"].shape
                adjusted_mask = np.zeros((2 * H_mask, 2 * W_mask), dtype=bool)
                adjusted_mask[
                    j * H_mask : (j + 1) * H_mask,
                    i * W_mask : (i + 1) * W_mask,
                ] = item["mask"]
                color_mask[
                    ["RED", "GREEN", "BLUE"].index(item["color"].upper()),
                    j * H_mask : (j + 1) * H_mask,
                    i * W_mask : (i + 1) * W_mask,
                ] = item["mask"]

                # dummy array to simulate the array field in the annotation
                array = np.random.rand(2, 2)
                array_path = tempdir / f"{array_counter}.npy"
                np.save(array_path, array)
                array_counter += 1

                adjusted_annotation = {
                    "class": item["vehicle_type"],
                    "instance_id": item["instance_id"],
                    "boundingbox": {
                        "x": x_new,
                        "y": y_new,
                        "w": w_new,
                        "h": h_new,
                    },
                    "array": {"path": array_path},
                    "keypoints": {"keypoints": item["keypoints"]},
                    "instance_segmentation": {"mask": adjusted_mask},
                    "metadata": {
                        "color": item["color"],
                        "brand": item["brand"],
                    },
                }
                annotations_list.append(adjusted_annotation)

            combined_filename = (
                "_".join([str(item["instance_id"]) for item in batch]) + ".jpg"
            )
            combined_filepath = tempdir / combined_filename
            cv2.imwrite(str(combined_filepath), combined_image)

            for annotation in annotations_list:
                yield {
                    "file": combined_filepath,
                    "task_name": annotation["class"],
                    "annotation": annotation,
                }
            for i, color in enumerate(["red", "green", "blue"]):
                yield {
                    "file": combined_filepath,
                    "task_name": "color",
                    "annotation": {
                        "class": color,
                        "segmentation": {"mask": color_mask[i]},
                    },
                }

            batch = []



from luxonis_ml.nn_archive.config_building_blocks import Head
from luxonis_ml.nn_archive.config_building_blocks.base_models.head_metadata import (
    HeadClassificationMetadata,
    HeadMetadata,
    HeadObjectDetectionMetadata,
    HeadObjectDetectionSSDMetadata,
    HeadSegmentationMetadata,
    HeadYOLOMetadata,
)

head_metadata = HeadMetadata(postprocessor_path="postprocessor.onnx")

head_object_detection_metadata = HeadObjectDetectionMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    postprocessor_path=None,
)

head_classification_metadata = HeadClassificationMetadata(
    classes=["person", "car"],
    n_classes=2,
    is_softmax=True,
    postprocessor_path="postprocessor.onnx",
)

head_segmentation_metadata = HeadSegmentationMetadata(
    classes=["person", "car"],
    n_classes=2,
    is_softmax=True,
    postprocessor_path=None,
)

head_object_detection_ssd_metadata = HeadObjectDetectionSSDMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    postprocessor_path=None,
    boxes_outputs="boxes",
    scores_outputs="scores",
)

head_yolo_obj_det_metadata = HeadYOLOMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    n_prototypes=None,
    n_keypoints=None,
    is_softmax=None,
    subtype="yolov6",
    postprocessor_path=None,
    yolo_outputs=["feats"],
    mask_outputs=None,
    protos_outputs=None,
    keypoints_outputs=None,
    angles_outputs=None,
)

head_yolo_instance_seg_metadata = HeadYOLOMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    n_prototypes=10,
    n_keypoints=None,
    is_softmax=True,
    subtype="yolov6",
    postprocessor_path="postprocessor.onnx",
    yolo_outputs=["feats"],
    mask_outputs=["mask"],
    protos_outputs="protos",
    keypoints_outputs=None,
    angles_outputs=None,
)

head_yolo_keypoint_det_metadata = HeadYOLOMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    n_prototypes=None,
    n_keypoints=21,
    is_softmax=None,
    subtype="yolov6",
    postprocessor_path=None,
    yolo_outputs=["feats"],
    mask_outputs=None,
    protos_outputs=None,
    keypoints_outputs=["keypoints"],
    angles_outputs=None,
)

head_yolo_obb_det_metadata = HeadYOLOMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    n_prototypes=None,
    n_keypoints=None,
    is_softmax=None,
    subtype="yolov6",
    postprocessor_path="postprocessor.onnx",
    yolo_outputs=["feats"],
    mask_outputs=None,
    protos_outputs=None,
    keypoints_outputs=None,
    angles_outputs=["angles"],
)

head_yolo_instance_seg_kpts_metadata = HeadYOLOMetadata(
    classes=["person", "car"],
    n_classes=2,
    iou_threshold=0.5,
    conf_threshold=0.5,
    max_det=1000,
    anchors=None,
    n_prototypes=10,
    n_keypoints=21,
    is_softmax=False,
    subtype="yolov6",
    postprocessor_path="postprocessor.onnx",
    yolo_outputs=["feats"],
    mask_outputs=["mask"],
    protos_outputs="protos",
    keypoints_outputs=["keypoints"],
    angles_outputs=None,
)

head_segmentation_metadata = HeadSegmentationMetadata(
    classes=["person", "car"],
    n_classes=2,
    is_softmax=False,
    custom_field="custom_field",  # type: ignore
    postprocessor_path="postprocessor.onnx",
)

classification_head = dict(
    Head(
        name="ClassificationHead",
        parser="Classification",
        outputs=["output"],
        metadata=head_classification_metadata,
    )
)

ssd_object_detection_head = dict(
    Head(
        name="ObjectDetectionSSDHead",
        parser="ObjectDetectionSSD",
        outputs=["boxes"],
        metadata=head_object_detection_ssd_metadata,
    )
)

yolo_object_detection_head = dict(
    Head(
        name="YoloDetectionHead",
        parser="YOLO",
        outputs=["output"],
        metadata=head_yolo_obb_det_metadata,
    )
)

yolo_instance_segmentation_head = dict(
    Head(
        name="YoloInstanceSegHead",
        parser="YOLO",
        outputs=["output"],
        metadata=head_yolo_instance_seg_metadata,
    )
)

yolo_keypoint_detection_head = dict(
    Head(
        name="YoloKeypointDetectionHead",
        parser="YOLO",
        outputs=["output"],
        metadata=head_yolo_keypoint_det_metadata,
    )
)

yolo_obb_detection_head = dict(
    Head(
        name="YoloOBBHead",
        parser="YOLO",
        outputs=["output"],
        metadata=head_yolo_obb_det_metadata,
    )
)

yolo_instance_seg_kpts_head = dict(
    Head(
        name="YoloInstaceSegKptHead",
        parser="YOLO",
        outputs=["outputs"],
        metadata=head_yolo_instance_seg_kpts_metadata,
    )
)

custom_segmentation_head = dict(
    Head(
        name="SegmentationHead",
        parser="Segmentation",
        outputs=["output"],
        metadata=head_segmentation_metadata,
    )
)



from abc import ABC, abstractmethod
from typing import (
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from semver.version import Version
from typing_extensions import TypeAlias

from luxonis_ml.data.datasets.annotation import DatasetRecord
from luxonis_ml.data.datasets.source import LuxonisSource
from luxonis_ml.typing import PathType
from luxonis_ml.utils import AutoRegisterMeta, Registry

DATASETS_REGISTRY: Registry[Type["BaseDataset"]] = Registry(name="datasets")


DatasetIterator: TypeAlias = Iterator[Union[dict, DatasetRecord]]


class BaseDataset(
    ABC, metaclass=AutoRegisterMeta, registry=DATASETS_REGISTRY, register=False
):
    """Base abstract dataset class for managing datasets in the Luxonis
    MLOps ecosystem."""

    @property
    @abstractmethod
    def identifier(self) -> str:
        """The unique identifier for the dataset.

        @type: str
        """
        ...

    @property
    @abstractmethod
    def version(self) -> Version:
        """The version of the underlying LDF.

        @type: L{Version}
        """
        ...

    @abstractmethod
    def set_tasks(self, tasks: Dict[str, List[str]]) -> None:
        """Sets the tasks for the dataset.

        @type tasks: Dict[str, List[str]]
        @param tasks: A dictionary mapping task names to task types.
        """
        ...

    @abstractmethod
    def get_tasks(self) -> Dict[str, str]:
        """Returns a dictionary mapping task names to task types.

        @rtype: Dict[str, str]
        @return: A dictionary mapping task names to task types.
        """
        ...

    @abstractmethod
    def update_source(self, source: LuxonisSource) -> None:
        """Updates underlying source of the dataset with a new
        LuxonisSource.

        @type source: L{LuxonisSource}
        @param source: The new C{LuxonisSource} to replace the old one.
        """
        ...

    @abstractmethod
    def set_classes(
        self,
        classes: Union[List[str], Dict[str, int]],
        task: Optional[str] = None,
    ) -> None:
        """Sets the classes for the dataset. This can be across all CV
        tasks or certain tasks.

        @type classes: Union[List[str], Dict[str, int]]
        @param classes: Either a list of class names or a dictionary
            mapping class names to class IDs. If list is provided, the
            class IDs will be assigned I{alphabetically} starting from
            C{0}. If the class names contain the class C{"background"},
            it will be assigned the class ID C{0}.
        @type task: Optional[str]
        @param task: Optionally specify the task where these classes
            apply.
        """
        ...

    @abstractmethod
    def get_classes(self) -> Dict[str, List[str]]:
        """Get classes according to computer vision tasks.

        @rtype: Dict[str, List[str]]
        @return: A dictionary mapping tasks to the classes used in each
            task.
        """
        ...

    @abstractmethod
    def set_skeletons(
        self,
        labels: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
        task: Optional[str] = None,
    ) -> None:
        """Sets the semantic structure of keypoint skeletons for the
        classes that use keypoints.

        Example::

            dataset.set_skeletons(
                labels=["right hand", "right shoulder", ...],
                edges=[[0, 1], [4, 5], ...]
            )

        @type labels: Optional[List[str]]
        @param labels: List of keypoint names.
        @type edges: Optional[List[Tuple[int, int]]]
        @param edges: List of edges between keypoints.
        @type task: Optional[str]
        @param task: Optionally specify the task where these skeletons apply.
            If not specified, the skeletons are set for all tasks that use keypoints.
        """
        ...

    @abstractmethod
    def get_skeletons(
        self,
    ) -> Dict[str, Tuple[List[str], List[Tuple[int, int]]]]:
        """Returns the dictionary defining the semantic skeleton for
        each class using keypoints.

        @rtype: Dict[str, Tuple[List[str], List[Tuple[int, int]]]]
        @return: For each task, a tuple containing a list of keypoint
            names and a list of edges between the keypoints.
        """
        ...

    @abstractmethod
    def add(
        self, generator: DatasetIterator, batch_size: int = 1_000_000
    ) -> None:
        """Write annotations to parquet files.

        @type generator: L{DatasetIterator}
        @param generator: A Python iterator that yields either instances
            of C{DatasetRecord} or a dictionary that can be converted to
            C{DatasetRecord}.
        @type batch_size: int
        @param batch_size: The number of annotations generated before
            processing. This can be set to a lower value to reduce
            memory usage.
        """
        ...

    @abstractmethod
    def make_splits(
        self,
        splits: Optional[
            Union[
                Dict[str, Sequence[PathType]],
                Dict[str, float],
                Tuple[float, float, float],
            ]
        ] = None,
        *,
        ratios: Optional[
            Union[Dict[str, float], Tuple[float, float, float]]
        ] = None,
        definitions: Optional[Dict[str, List[PathType]]] = None,
        replace_old_splits: bool = False,
    ) -> None:
        """Generates splits for the dataset.

        @type splits: Optional[Union[Dict[str, Sequence[PathType]], Dict[str, float], Tuple[float, float, float]]]
        @param splits: A dictionary of splits or a tuple of ratios for train, val, and test splits. Can be one of:
            - A dictionary of splits with keys as split names and values as lists of filepaths
            - A dictionary of splits with keys as split names and values as ratios
            - A 3-tuple of ratios for train, val, and test splits
        @type ratios: Optional[Union[Dict[str, float], Tuple[float, float, float]]]
        @param ratios: Deprecated! A dictionary of splits with keys as split names and values as ratios.
        @type definitions: Optional[Dict[str, List[PathType]]]
        @param definitions: Deprecated! A dictionary of splits with keys as split names and values as lists of filepaths.
        @type replace_old_splits: bool
        @param replace_old_splits: Whether to remove old splits and generate new ones. If set to False, only new files will be added to the splits. Default is False.
        """
        ...

    @abstractmethod
    def delete_dataset(self) -> None:
        """Deletes all local files belonging to the dataset."""
        ...

    @staticmethod
    @abstractmethod
    def exists(dataset_name: str) -> bool:
        """Checks whether a dataset exists.

        @warning: For offline mode only.
        @type dataset_name: str
        @param dataset_name: Name of the dataset
        @rtype: bool
        @return: Whether the dataset exists
        """
        ...

    def get_task_names(self) -> List[str]:
        """Get the task names for the dataset.

        Like `get_tasks`, but returns only the task names
        instead of the entire names.

        @rtype: List[str]
        @return: List of task names.
        """
        return list(self.get_tasks().keys())

    def get_n_keypoints(self) -> Dict[str, int]:
        skeletons = self.get_skeletons()
        return {task: len(skeletons[task][0]) for task in skeletons}



"""Model Utility Functions.

This script provides utility functions for handling ONNX models.
It allows manipulating ONNX models in order to extract intermediate outputs aka embeddings.

Functions:
    - load_model_onnx(model_path: str = "resnet50.onnx") -> onnx.ModelProto:
        Loads an ONNX model from the specified file path.

    - save_model_onnx(model: onnx.ModelProto, model_path_out: str = "resnet50.onnx"):
        Saves an ONNX model to the specified file path.

    - extend_output_onnx(onnx_model: onnx.ModelProto, intermediate_tensor_name: str = "/Flatten_output_0") -> onnx.ModelProto:
        Sets an intermediate output layer as output of the provided ONNX model.
        If C{overwrite} is set to True, the second to last layer output will be set as output layer and renamed.

Dependencies:
    - onnx
"""

import onnx


def load_model_onnx(model_path: str = "resnet50.onnx") -> onnx.ModelProto:
    """Load an ONNX model from the provided path."""
    return onnx.load(model_path)


def save_model_onnx(
    model: onnx.ModelProto, model_path_out: str = "resnet50.onnx"
) -> None:
    """Save an ONNX model to the specified file path."""
    onnx.save(model, model_path_out)


def extend_output_onnx(
    onnx_model: onnx.ModelProto,
    intermediate_tensor_name: str = "/Flatten_output_0",
    overwrite: bool = False,
) -> onnx.ModelProto:
    """Set an intermediate output layer as output of the provided ONNX
    model.

    If C{overwrite} is set to True, the second to last layer output will
    be set as output layer and renamed.

    (You need to know the name of the intermediate layer, which you can
    find by inspecting the ONNX model with Netron.app)
    """
    if overwrite:
        onnx.checker.check_model(onnx_model)
        second_to_last_node = onnx_model.graph.node[-2]
        old_name = second_to_last_node.output[0]
        second_to_last_node.output[0] = intermediate_tensor_name

        for node in onnx_model.graph.node:
            for idx, input_name in enumerate(node.input):
                if input_name == old_name:
                    node.input[idx] = intermediate_tensor_name

    intermediate_layer_value_info = onnx.helper.ValueInfoProto()
    intermediate_layer_value_info.name = intermediate_tensor_name
    onnx_model.graph.output.extend([intermediate_layer_value_info])
    return onnx_model



from contextlib import suppress
from typing import List, Optional

from pydantic import Field, model_validator
from typing_extensions import Self

from luxonis_ml.nn_archive.config_building_blocks.enums import DataType
from luxonis_ml.nn_archive.utils import infer_layout
from luxonis_ml.utils import BaseModelExtraForbid


class Output(BaseModelExtraForbid):
    """Represents output stream of a model.

    @type name: str
    @ivar name: Name of the output layer.
    @type dtype: DataType
    @ivar dtype: Data type of the output data (e.g., 'float32').
    """

    name: str = Field(description="Name of the output layer.")
    dtype: DataType = Field(
        description="Data type of the output data (e.g., 'float32')."
    )
    shape: Optional[List[int]] = Field(
        None,
        description="Shape of the output as a list of integers (e.g. [1, 1000]).",
    )
    layout: Optional[str] = Field(
        None,
        description="List of letters describing the output layout (e.g. 'NC').",
    )

    @model_validator(mode="after")
    def validate_layout(self) -> Self:
        if self.layout is None:
            return self

        self.layout = self.layout.upper()

        if "N" in self.layout and self.layout[0] != "N":
            raise ValueError(
                "If N (batch size) is included in the layout, it must be first"
            )

        if self.shape is None:
            raise ValueError("Shape must be defined if layout is defined.")

        if len(self.layout) != len(self.shape):
            raise ValueError("Layout and shape must have the same length.")

        return self

    @model_validator(mode="after")
    def infer_layout(self) -> Self:
        if self.layout is None and self.shape is not None:
            with suppress(Exception):
                self.layout = infer_layout(self.shape)

        return self



import numpy as np

from luxonis_ml.data import AlbumentationsEngine


def test_metadata_no_boxes():
    config = [
        {
            "name": "Defocus",
            "params": {"p": 1.0},
        },
        {
            "name": "Mosaic4",
            "params": {"p": 1.0, "out_width": 640, "out_height": 640},
        },
    ]
    augmentations = AlbumentationsEngine(
        256, 256, {"/metadata/id": "metadata/id"}, {"/metadata/id": 0}, config
    )
    _, labels = augmentations.apply(
        [
            (np.zeros((3, 256, 256)), {"/metadata/id": np.array([i])})
            for i in range(4)
        ]
    )
    assert labels["/metadata/id"].tolist() == [0, 1, 2, 3]


def test_skip_augmentations():
    config = [
        {
            "name": "Perspective",
        },
        {
            "name": "Flip",
        },
        {
            "name": "HorizontalFlip",
        },
        {
            "name": "VerticalFlip",
        },
        {
            "name": "Rotate",
        },
        {
            "name": "Mosaic4",
            "params": {"out_width": 640, "out_height": 640},
        },
    ]
    targets = {
        "/boundingbox": "boundingbox",
        "/classification": "classification",
        "/keypoints": "keypoints",
        "/instance_segmentation": "instance_segmentation",
        "/segmentation": "segmentation",
    }
    n_classes = {
        "/boundingbox": 1,
        "/classification": 1,
        "/keypoints": 1,
        "/instance_segmentation": 1,
        "/segmentation": 1,
    }
    augmentations = AlbumentationsEngine(256, 256, targets, n_classes, config)

    spatial_transform_names = next(
        (
            [t.__class__.__name__ for t in cell.cell_contents.transforms]
            for cell in augmentations.spatial_transform.__closure__  # type: ignore
            if hasattr(cell.cell_contents, "transforms")
        ),
        [],
    )

    batched_transform_names = [
        t.__class__.__name__ for t in augmentations.batch_transform.transforms
    ]

    assert spatial_transform_names == ["Rotate"]
    assert batched_transform_names == ["Mosaic4"]



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class DarknetParser(BaseParser):
    """Parses directory with DarkNet annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img1.txt
        │   ├── ...
        │   └── _darknet.labels
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        if not (split_path / "_darknet.labels").exists():
            return None
        images = BaseParser._list_images(split_path)
        labels = split_path.glob("*.txt")
        if not BaseParser._compare_stem_files(images, labels):
            return None
        return {
            "image_dir": split_path,
            "classes_path": split_path / "_darknet.labels",
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if DarknetParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            classes_path=dataset_dir / "train" / "_darknet.labels",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            classes_path=dataset_dir / "valid" / "_darknet.labels",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test",
            classes_path=dataset_dir / "test" / "_darknet.labels",
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(self, image_dir: Path, classes_path: Path) -> ParserOutput:
        """Parses annotations from Darknet format to LDF. Annotations
        include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type classes_path: Path
        @param classes_path: Path to file with class names
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """
        with open(classes_path) as f:
            class_names = {
                i: line.rstrip() for i, line in enumerate(f.readlines())
            }

        def generator() -> DatasetIterator:
            for img_path in self._list_images(image_dir):
                ann_path = img_path.with_suffix(".txt")
                file = str(img_path)
                with open(ann_path) as f:
                    annotation_data = f.readlines()

                for ann_line in annotation_data:
                    class_id, x_center, y_center, width, height = list(
                        ann_line.split(" ")
                    )
                    class_name = class_names[int(class_id)]

                    yield {
                        "file": file,
                        "annotation": {
                            "class": class_name,
                            "boundingbox": {
                                "x": float(x_center) - float(width) / 2,
                                "y": float(y_center) - float(height) / 2,
                                "w": float(width),
                                "h": float(height),
                            },
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), {}, added_images
